import dataclasses as dc
import typing as ta

from ... import lang
from ... import reflect as rfl
from ...funcs import match as mfs
from .contexts import MarshalContext
from .contexts import UnmarshalContext
from .types import Marshaler
from .types import MarshalerFactory
from .types import MarshalerMaker
from .types import Unmarshaler
from .types import UnmarshalerFactory
from .types import UnmarshalerMaker
from .values import Value


##


@dc.dataclass(frozen=True)
class FuncMarshaler(Marshaler, lang.Final):
    fn: ta.Callable[[MarshalContext, ta.Any], Value]

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        return self.fn(ctx, o)


@dc.dataclass(frozen=True)
class FuncUnmarshaler(Unmarshaler, lang.Final):
    fn: ta.Callable[[UnmarshalContext, Value], ta.Any]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return self.fn(ctx, v)


##


@dc.dataclass(frozen=True)
class FuncMarshalerFactory(MarshalerFactory):  # noqa
    guard: ta.Callable[[MarshalContext, rfl.Type], bool]
    fn: ta.Callable[[MarshalContext, rfl.Type], Marshaler]

    @lang.cached_property
    def make_marshaler(self) -> MarshalerMaker:
        return mfs.simple(self.guard, self.fn)


@dc.dataclass(frozen=True)
class FuncUnmarshalerFactory(UnmarshalerFactory):  # noqa
    guard: ta.Callable[[UnmarshalContext, rfl.Type], bool]
    fn: ta.Callable[[UnmarshalContext, rfl.Type], Unmarshaler]

    @lang.cached_property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return mfs.simple(self.guard, self.fn)



"""
FIXME:
 - lol support late registration and such, this is broken
  - need to cache in the registry somehow, cannot iterate over all ModuleImports each time
   - probably use something like dataclass driver context items or some such

TODO:
 - per-type module imports
"""
import threading
import typing as ta

from .... import reflect as rfl
from ....funcs import match as mfs
from ...base.contexts import BaseContext
from ...base.contexts import MarshalContext
from ...base.contexts import UnmarshalContext
from ...base.types import Marshaler
from ...base.types import MarshalerFactory
from ...base.types import MarshalerMaker
from ...base.types import Unmarshaler
from ...base.types import UnmarshalerFactory
from ...base.types import UnmarshalerMaker
from .configs import ModuleImport


R = ta.TypeVar('R')
ContextT = ta.TypeVar('ContextT', bound=BaseContext)


##


class _ModuleImportingFactory(mfs.MatchFn[[ContextT, rfl.Type], R]):
    def __init__(
            self,
            f: mfs.MatchFn[[ContextT, rfl.Type], R],
            callback: ta.Callable[[], None] | None = None,
    ) -> None:
        super().__init__()

        self._f = f
        self._callback = callback

        self._lock = threading.RLock()
        self._has_imported = False

    def _do_import(self, ctx: ContextT) -> None:
        c = 0
        for mi in ctx.config_registry.get_of(None, ModuleImport):
            if mi.import_if_necessary():
                c += 1

        if c:
            if self._callback is not None:
                self._callback()

    def _import_if_necessary(self, ctx: ContextT) -> None:
        # FIXME:
        # if not self._has_imported:
        #     with self._lock:
        #         if not self._has_imported:
        #             self._do_import(ctx)
        #             self._has_imported = True
        self._do_import(ctx)

    def guard(self, ctx: ContextT, rty: rfl.Type) -> bool:
        self._import_if_necessary(ctx)
        return self._f.guard(ctx, rty)

    def fn(self, ctx: ContextT, rty: rfl.Type) -> R:
        self._import_if_necessary(ctx)
        return self._f(ctx, rty)


##


class ModuleImportingMarshalerFactory(MarshalerFactory):
    def __init__(
            self,
            f: MarshalerFactory,
            callback: ta.Callable[[], None] | None = None,
    ) -> None:
        super().__init__()

        self._f = f
        self._tcf: _ModuleImportingFactory[MarshalContext, Marshaler] = _ModuleImportingFactory(
            f.make_marshaler,
            callback,
        )

    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self._tcf


class ModuleImportingUnmarshalerFactory(UnmarshalerFactory):
    def __init__(
            self,
            f: UnmarshalerFactory,
            callback: ta.Callable[[], None] | None = None,
    ) -> None:
        super().__init__()

        self._f = f
        self._tcf: _ModuleImportingFactory[UnmarshalContext, Unmarshaler] = _ModuleImportingFactory(
            f.make_unmarshaler,
            callback,
        )

    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self._tcf



import dataclasses as dc
import importlib
import sys

from .... import lang
from ...base.configs import Config


##


@dc.dataclass(frozen=True, eq=False)
class ModuleImport(Config, lang.Final):
    name: str
    package: str | None = None

    @lang.cached_function
    def resolve(self) -> str:
        return lang.resolve_import_name(self.name, self.package)

    def import_if_necessary(self) -> bool:
        if (mn := self.resolve()) in sys.modules:
            return False

        importlib.import_module(mn)
        return True



"""
~> https://github.com/google/guice/commit/70248eafa90cd70a68b293763e53f6aec656e73c
"""
from .. import dataclasses as _dc


_dc.init_package(
    globals(),
    codegen=True,
)


##


from .. import lang as _lang  # noqa


with _lang.auto_proxy_init(globals()):
    ##

    from .binder import (  # noqa
        bind,
        bind_as_fn,
        bind_map_entry_const,
        bind_set_entry_const,
    )

    from .bindings import (  # noqa
        Binding,
    )

    from .eagers import (  # noqa
        Eager,
    )

    from .elements import (  # noqa
        Element,
        Elemental,
        Elements,
        as_elements,
    )

    from .errors import (  # noqa
        BaseKeyError,
        ConflictingKeyError,
        CyclicDependencyError,
        ScopeAlreadyOpenError,
        ScopeError,
        ScopeNotOpenError,
        UnboundKeyError,
    )

    from .injector import (  # noqa
        AsyncInjector,
        create_async_injector,
    )

    from .inspect import (  # noqa
        Kwarg,
        KwargsTarget,
        build_kwargs_target,
        tag,
    )

    from .keys import (  # noqa
        Key,
        as_key,
    )

    from .listeners import (  # noqa
        ProvisionListener,
        ProvisionListenerBinding,
        bind_provision_listener,
    )

    from .managed import (  # noqa
        create_async_managed_injector,
        make_async_managed_provider,

        create_managed_injector,
        make_managed_provider,

        create_maysync_managed_injector,
        make_maysync_managed_provider,
    )

    from .maysync import (  # noqa
        MaysyncInjector,
        create_maysync_injector,
    )

    from .multis import (  # noqa
        MapBinding,
        MapProvider,
        SetBinding,
        SetProvider,
        MapBinder,
        SetBinder,

        MapBinder as map_binder,  # noqa
        SetBinder as set_binder,  # noqa
    )


    from .overrides import (  # noqa
        Overrides,
        override,
    )

    from .origins import (  # noqa
        HasOrigins,
        Origin,
        Origins,
    )

    from .privates import (  # noqa
        Expose,
        Private,
        private,

        Expose as expose,  # noqa
    )

    from .providers import (  # noqa
        AsyncFnProvider,
        ConstProvider,
        CtorProvider,
        FnProvider,
        LinkProvider,
        Provider,
    )

    from .scopes import (  # noqa
        ScopeBinding,
        bind_scope,

        Singleton,

        ThreadScope,

        SeededScope,
        ScopeSeededProvider,
        bind_scope_seed,

        async_enter_seeded_scope,
        enter_seeded_scope,
        maysync_enter_seeded_scope,
    )

    from .sync import (  # noqa
        Injector,
        create_injector,
    )

    from .tags import (  # noqa
        Id,
    )

    from .types import (  # noqa
        Scope,
        Tag,
        Unscoped,
    )

    from .utils import (  # noqa
        ConstFn,
    )



from ..lite.maybes import Maybe


##

# These are classmethods on the lite `Maybe` class, and thus can't be directly `auto_proxy_init` imported in lang's
# `__init__.py`.
empty = Maybe.empty
just = Maybe.just



"""
TODO:
 - if already imported just return?
"""
import contextlib
import functools
import importlib.util
import types
import typing as ta

from ..lazyglobals import LazyGlobals
from .capture import ImportCapture
from .capture import _new_import_capture_hook


##


def proxy_import(
        spec: str,
        package: str | None = None,
        extras: ta.Iterable[str] | None = None,
) -> types.ModuleType:
    if isinstance(extras, str):
        raise TypeError(extras)

    omod = None

    def __getattr__(att):  # noqa
        nonlocal omod
        if omod is None:
            omod = importlib.import_module(spec, package=package)
            if extras:
                for x in extras:
                    importlib.import_module(f'{spec}.{x}', package=package)
        return getattr(omod, att)

    lmod = types.ModuleType(spec)
    lmod.__getattr__ = __getattr__  # type: ignore[method-assign]
    return lmod


#


def auto_proxy_import(
        mod_globals: ta.MutableMapping[str, ta.Any],
        *,
        disable: bool = False,

        unreferenced_callback: ta.Callable[[ta.Mapping[str, ta.Sequence[str | None]]], None] | None = None,
        raise_unreferenced: bool = False,

        _stack_offset: int = 0,
) -> ta.ContextManager[ImportCapture]:
    inst = ImportCapture(
        mod_globals,
        _hook=_new_import_capture_hook(
            mod_globals,
            stack_offset=_stack_offset + 1,
        ),
        disable=disable,
    )

    @contextlib.contextmanager
    def inner() -> ta.Iterator[ImportCapture]:
        with inst.capture(
                unreferenced_callback=unreferenced_callback,
                raise_unreferenced=raise_unreferenced,
        ):
            yield inst

        pkg = mod_globals.get('__package__')
        for pi in inst.captured.imports:
            for sa, ma in pi.attrs:
                mod_globals[ma] = proxy_import(pi.spec + (('.' + sa) if sa is not None else ''), pkg)

    return inner()


##


class _ProxyInit:
    class NamePackage(ta.NamedTuple):
        name: str
        package: str

    class _Import(ta.NamedTuple):
        spec: str
        attr: str | None

    def __init__(
            self,
            lazy_globals: LazyGlobals,
            name_package: NamePackage,
    ) -> None:
        super().__init__()

        self._lazy_globals = lazy_globals
        self._name_package = name_package

        self._imps_by_attr: dict[str, _ProxyInit._Import] = {}
        self._mods_by_spec: dict[str, ta.Any] = {}

    @property
    def name_package(self) -> NamePackage:
        return self._name_package

    def add(
            self,
            package: str,
            attrs: ta.Iterable[tuple[str | None, str]],
    ) -> None:
        for imp_attr, as_attr in attrs:
            if imp_attr is None:
                self._imps_by_attr[as_attr] = self._Import(package, None)
                self._lazy_globals.set_fn(as_attr, functools.partial(self.get, as_attr))

            else:
                self._imps_by_attr[as_attr] = self._Import(package, imp_attr)
                self._lazy_globals.set_fn(as_attr, functools.partial(self.get, as_attr))

    def _import_module(self, name: str) -> ta.Any:
        return importlib.import_module(name, package=self._name_package.package)

    def get(self, attr: str) -> ta.Any:
        try:
            imp = self._imps_by_attr[attr]
        except KeyError:
            raise AttributeError(attr)  # noqa

        val: ta.Any

        if imp.attr is None:
            val = self._import_module(imp.spec)

        else:
            try:
                mod = self._mods_by_spec[imp.spec]
            except KeyError:
                mod = self._import_module(imp.spec)
                self._mods_by_spec[imp.spec] = mod

            val = getattr(mod, imp.attr)

        return val


def proxy_init(
        init_globals: ta.MutableMapping[str, ta.Any],
        spec: str,
        attrs: ta.Iterable[str | tuple[str | None, str | None] | None] | None = None,
) -> None:
    if isinstance(attrs, str):
        raise TypeError(attrs)

    if attrs is None:
        attrs = [None]

    whole_attr = spec.split('.')[-1]
    al: list[tuple[str | None, str]] = []
    for attr in attrs:
        if attr is None:
            al.append((None, whole_attr))

        elif isinstance(attr, str):
            al.append((attr, attr))

        elif isinstance(attr, tuple):
            imp_attr, as_attr = attr
            if as_attr is None:
                if imp_attr is None:
                    as_attr = whole_attr
                else:
                    as_attr = imp_attr
            al.append((imp_attr, as_attr))

        else:
            raise TypeError(attr)

    #

    init_name_package = _ProxyInit.NamePackage(
        init_globals['__name__'],
        init_globals['__package__'],
    )

    pi: _ProxyInit
    try:
        pi = init_globals['__proxy_init__']

    except KeyError:
        pi = _ProxyInit(
            LazyGlobals.install(init_globals),
            init_name_package,
        )
        init_globals['__proxy_init__'] = pi

    else:
        if pi.name_package != init_name_package:
            raise Exception(f'Wrong init name: {pi.name_package=} != {init_name_package=}')

    pi.add(spec, al)


#


def auto_proxy_init(
        init_globals: ta.MutableMapping[str, ta.Any],
        *,
        disable: bool = False,
        eager: bool = False,

        unreferenced_callback: ta.Callable[[ta.Mapping[str, ta.Sequence[str | None]]], None] | None = None,
        raise_unreferenced: bool = False,

        update_exports: bool = False,

        _stack_offset: int = 0,
) -> ta.ContextManager[ImportCapture]:
    inst = ImportCapture(
        init_globals,
        _hook=_new_import_capture_hook(
            init_globals,
            stack_offset=_stack_offset + 1,
        ),
        disable=disable,
    )

    @contextlib.contextmanager
    def inner() -> ta.Iterator[ImportCapture]:
        with inst.capture(
                unreferenced_callback=unreferenced_callback,
                raise_unreferenced=raise_unreferenced,
        ):
            yield inst

        for pi in inst.captured.imports:
            proxy_init(
                init_globals,
                pi.spec,
                pi.attrs,
            )

        if eager:
            lg = LazyGlobals.install(init_globals)

            for a in inst.captured.attrs:
                lg.get(a)

        if update_exports:
            inst.update_exports()

    return inner()



import typing as ta


##


class AmbiguousLazyGlobalsFallbackError(Exception):
    def __init__(self, attr: str, fallbacks: list[ta.Callable[[str], ta.Any]]) -> None:
        super().__init__()

        self.attr = attr
        self.fallbacks = fallbacks

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.attr!r}, {self.fallbacks!r})'


class LazyGlobals:
    def __init__(
            self,
            *,
            globals: ta.MutableMapping[str, ta.Any] | None = None,  # noqa
            update_globals: bool = False,
    ) -> None:
        super().__init__()

        self._globals = globals
        self._update_globals = update_globals

        self._attr_fns: dict[str, ta.Callable[[], ta.Any]] = {}
        self._fallback_fns: list[ta.Callable[[str], ta.Callable[[], ta.Any]]] = []

    @classmethod
    def install(cls, globals: ta.MutableMapping[str, ta.Any]) -> 'LazyGlobals':  # noqa
        try:
            xga = globals['__getattr__']
        except KeyError:
            pass
        else:
            if not isinstance(xga, cls):
                raise RuntimeError(f'Module already has __getattr__ hook: {xga}')  # noqa
            return xga

        lm = cls(
            globals=globals,
            update_globals=True,
        )

        globals['__getattr__'] = lm

        return lm

    def set_fn(self, attr: str, fn: ta.Callable[[], ta.Any]) -> 'LazyGlobals':
        self._attr_fns[attr] = fn
        return self

    def add_fallback_fn(self, fn: ta.Callable[[str], ta.Callable[[], ta.Any]]) -> 'LazyGlobals':
        self._fallback_fns.append(fn)
        return self

    def get(self, attr: str) -> ta.Any:
        val: ta.Any

        if (attr_fn := self._attr_fns.get(attr)) is not None:
            val = attr_fn()

        elif (fallbacks := [(fb_fn, fb_fn(attr)) for fb_fn in self._fallback_fns]):
            if len(fallbacks) > 1:
                raise AmbiguousLazyGlobalsFallbackError(attr, [fb_fn for fb_fn, _ in fallbacks])
            [val] = fallbacks

        else:
            raise AttributeError(attr)

        if self._update_globals and self._globals is not None:
            self._globals[attr] = val

        return val

    def __call__(self, attr: str) -> ta.Any:
        return self.get(attr)



import contextlib
import contextvars
import typing as ta

from .... import lang
from ...specs import ClassSpec
from .. import concerns as _concerns  # noqa  # imported for registration
from ..configs import DEFAULT_PACKAGE_CONFIG
from ..configs import PACKAGE_CONFIG_CACHE
from ..generation import processor as gp
from .base import ProcessingContext
from .base import ProcessingOption
from .base import Processor
from .registry import all_processing_context_item_factories
from .registry import ordered_processor_types


##


_OPTIONS_CONTEXT_VAR: contextvars.ContextVar[ta.Sequence[ProcessingOption]] = contextvars.ContextVar(
    f'{__name__}._OPTIONS_CONTEXT_VAR',
    default=(),
)


@contextlib.contextmanager
def processing_options_context(*opts: ProcessingOption) -> ta.Iterator[None]:
    with lang.context_var_setting(
            _OPTIONS_CONTEXT_VAR,
            (*_OPTIONS_CONTEXT_VAR.get(), *opts),
    ):
        yield


##


def drive_cls_processing(
        cls: type,
        cs: ClassSpec,
        *,
        plan_only: bool = False,
        verbose: bool = False,
) -> type:
    options: list[ProcessingOption] = list(_OPTIONS_CONTEXT_VAR.get())
    if plan_only:
        options.append(gp.PlanOnly(True))
    if verbose:
        options.append(gp.Verbose(True))

    #

    pkg_config = DEFAULT_PACKAGE_CONFIG
    cls_mod = cls.__module__
    if '.' in cls_mod:
        cls_pkg = cls_mod.rpartition('.')[0]
        pkg_cfg = PACKAGE_CONFIG_CACHE.get(cls_pkg, pkg_config)  # noqa

    #

    ctx = ProcessingContext(
        cls,
        cs,
        all_processing_context_item_factories(),
        options=options,
    )

    processors: list[Processor] = [
        proc_cls(ctx)
        for proc_cls in ordered_processor_types()
    ]

    for p in processors:
        p.check()

    ret = cls
    for p in processors:
        ret = p.process(ret)

    return ret



import typing as ta

from ..base.types import MarshalerFactory
from ..base.types import UnmarshalerFactory
from .marshal import PolymorphismMarshalerFactory
from .metadata import Polymorphism
from .metadata import TypeTagging
from .metadata import WrapperTypeTagging
from .unmarshal import PolymorphismUnmarshalerFactory


##


def standard_polymorphism_factories(
        poly: Polymorphism,
        tt: TypeTagging = WrapperTypeTagging(),
) -> ta.Sequence[MarshalerFactory | UnmarshalerFactory]:
    return [
        PolymorphismMarshalerFactory(poly, tt),
        PolymorphismUnmarshalerFactory(poly, tt),
    ]



from ... import check
from ... import reflect as rfl
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..factories.simple import SimpleMarshalerFactory
from ..factories.simple import SimpleUnmarshalerFactory


##


class NewtypeMarshalerFactory(SimpleMarshalerFactory):
    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.NewType)

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        return ctx.make(check.isinstance(rty, rfl.NewType).ty)


class NewtypeUnmarshalerFactory(SimpleUnmarshalerFactory):
    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.NewType)

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        return ctx.make(check.isinstance(rty, rfl.NewType).ty)



import typing as ta

from ... import check
from ... import lang
from ... import reflect as rfl
from ...funcs import match as mfs
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import MarshalerFactory
from ..base.types import MarshalerMaker
from ..base.types import Unmarshaler
from ..base.types import UnmarshalerFactory
from ..base.types import UnmarshalerMaker
from ..base.values import Value


T = ta.TypeVar('T')
R = ta.TypeVar('R')
C = ta.TypeVar('C')


##


class _RecursiveTypeFactory(mfs.MatchFn[[C, rfl.Type], R]):
    def __init__(
            self,
            f: mfs.MatchFn[[C, rfl.Type], R],
            prx: ta.Callable[[], tuple[R, ta.Callable[[R], None]]],
    ) -> None:
        super().__init__()

        self._f = f
        self._prx = prx
        self._dct: dict[rfl.Type, R] = {}

    def guard(self, ctx: C, rty: rfl.Type) -> bool:
        check.isinstance(rty, rfl.TYPES)
        return self._f.guard(ctx, rty)

    def fn(self, ctx: C, rty: rfl.Type) -> R:
        check.isinstance(rty, rfl.TYPES)
        try:
            return self._dct[rty]
        except KeyError:
            pass
        p, sp = self._prx()
        self._dct[rty] = p
        try:
            r = self._f(ctx, rty)
            sp(r)
            return r
        finally:
            del self._dct[rty]


##


class _Proxy(ta.Generic[T]):
    __obj: T | None = None

    @property
    def _obj(self) -> T:
        if self.__obj is None:
            raise TypeError('recursive proxy not set')
        return self.__obj

    def _set_obj(self, obj: T) -> None:
        if self.__obj is not None:
            raise TypeError('recursive proxy already set')
        self.__obj = obj

    @classmethod
    def _new(cls) -> tuple[ta.Any, ta.Callable[[ta.Any], None]]:
        return (p := cls()), p._set_obj  # noqa


##


class _ProxyMarshaler(_Proxy[Marshaler], Marshaler):
    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        return self._obj.marshal(ctx, o)


class RecursiveMarshalerFactory(MarshalerFactory, lang.Final):
    def __init__(self, f: MarshalerFactory) -> None:
        super().__init__()

        self._f = f
        self._rtf: _RecursiveTypeFactory[MarshalContext, Marshaler] = _RecursiveTypeFactory(
            self._f.make_marshaler,  # noqa
            _ProxyMarshaler._new,  # noqa
        )

    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self._rtf


class _ProxyUnmarshaler(_Proxy[Unmarshaler], Unmarshaler):
    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return self._obj.unmarshal(ctx, v)


class RecursiveUnmarshalerFactory(UnmarshalerFactory, lang.Final):
    def __init__(self, f: UnmarshalerFactory) -> None:
        super().__init__()

        self._f = f
        self._rtf: _RecursiveTypeFactory[UnmarshalContext, Unmarshaler] = _RecursiveTypeFactory(
            self._f.make_unmarshaler,  # noqa
            _ProxyUnmarshaler._new,  # noqa
        )

    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self._rtf



from .abstract import (  # noqa
    AbstractLifecycle,

    AbstractAsyncLifecycle,
)

from .base import (  # noqa
    AnyLifecycleCallback,
    AnyLifecycle,
    AnyCallbackLifecycle,

    LifecycleCallback,
    CallbackLifecycle,
    Lifecycle,

    AsyncLifecycleCallback,
    CallbackAsyncLifecycle,
    AsyncLifecycle,
)

from .contextmanagers import (  # noqa
    ContextManagerLifecycle,
    LifecycleContextManager,
)

from .controller import (  # noqa
    AnyLifecycleListener,
    AnyLifecycleController,

    LifecycleController,
    LifecycleListener,
)

from .manager import (  # noqa
    LifecycleManager,
)

from .states import (  # noqa
    LifecycleState,
    LifecycleStateError,
    LifecycleStates,
)

from .transitions import (  # noqa
    LifecycleTransition,
    LifecycleTransitions,
)



"""
TODO:
 - add origin to Id
"""
import itertools
import typing as ta

from ... import cached
from ... import check
from ... import dataclasses as dc
from ... import lang
from ..bindings import Binding
from ..eagers import Eager
from ..elements import Element
from ..injector import AsyncInjector
from ..keys import Key
from ..privates import Expose
from ..privates import Private
from ..providers import Provider
from ..scopes import Singleton
from .elements import ElementCollection
from .injector import AsyncInjectorImpl
from .providers import InternalProvider
from .providers import ProviderImpl


##


_PRIVATE_COUNT = itertools.count()


@dc.dataclass(frozen=True)
class PrivateInjectorId(lang.Final):
    id: int


##


@dc.dataclass(eq=False)
class PrivateInjectorProviderImpl(ProviderImpl, lang.Final):
    id: PrivateInjectorId
    ec: ElementCollection

    @property
    def providers(self) -> ta.Iterable[Provider]:
        return ()

    async def provide(self, injector: AsyncInjector) -> ta.Any:
        return check.isinstance(injector, AsyncInjectorImpl).create_child(self.ec)


##


@dc.dataclass(eq=False)
class ExposedPrivateProviderImpl(ProviderImpl, lang.Final):
    pik: Key
    k: Key

    @property
    def providers(self) -> ta.Iterable[Provider]:
        return ()

    async def provide(self, injector: AsyncInjector) -> ta.Any:
        pi = await injector.provide(self.pik)
        return await pi.provide(self.k)


##


@dc.dataclass(frozen=True)
class PrivateInfo(lang.Final):
    owner: ElementCollection
    p: Private

    @cached.property
    def id(self) -> PrivateInjectorId:
        return PrivateInjectorId(next(_PRIVATE_COUNT))

    @cached.property
    def pik(self) -> Key:
        return Key(AsyncInjectorImpl, tag=self.id)

    @cached.function
    def element_collection(self) -> ElementCollection:
        return ElementCollection(self.p.elements)

    ##

    @cached.function
    def private_provider_impl(self) -> PrivateInjectorProviderImpl:
        return PrivateInjectorProviderImpl(self.id, self.element_collection())

    @cached.function
    def exposed_provider_impls(self) -> ta.Sequence[ExposedPrivateProviderImpl]:
        exs = self.element_collection().elements_of_type(Expose)
        return [ExposedPrivateProviderImpl(self.pik, ex.key) for ex in exs]

    @cached.function
    def owner_elements(self) -> ta.Iterable[Element]:
        lst: list[Element] = [
            Binding(self.pik, InternalProvider(self.private_provider_impl()), Singleton()),
            Eager(self.pik),
            *(Binding(ep.k, InternalProvider(ep)) for ep in self.exposed_provider_impls()),
        ]
        return lst



# ruff: noqa: UP006 UP007 UP045 UP046
# @omlish-lite
import abc
import time
import typing as ta

from ..lite.abstract import Abstract
from .infos import LoggingContextInfo
from .infos import LoggingContextInfos
from .infos import LoggingExcInfoArg
from .infos import LoggingMsgFn
from .levels import LogLevel


LoggingContextInfoT = ta.TypeVar('LoggingContextInfoT', bound=LoggingContextInfo)


##


class LoggingContext(Abstract):
    @abc.abstractmethod
    def get_info(self, ty: ta.Type[LoggingContextInfoT]) -> ta.Optional[LoggingContextInfoT]:
        raise NotImplementedError

    @ta.final
    def __getitem__(self, ty: ta.Type[LoggingContextInfoT]) -> ta.Optional[LoggingContextInfoT]:
        return self.get_info(ty)

    @ta.final
    def must_get_info(self, ty: ta.Type[LoggingContextInfoT]) -> LoggingContextInfoT:
        if (info := self.get_info(ty)) is None:
            raise TypeError(f'LoggingContextInfo absent: {ty}')
        return info


@ta.final
class SimpleLoggingContext(LoggingContext):
    def __init__(self, *infos: LoggingContextInfo) -> None:
        self._infos: ta.Dict[ta.Type[LoggingContextInfo], LoggingContextInfo] = {type(i): i for i in infos}

    def get_info(self, ty: ta.Type[LoggingContextInfoT]) -> ta.Optional[LoggingContextInfoT]:
        return self._infos.get(ty)


##


class CaptureLoggingContext(LoggingContext, Abstract):
    @abc.abstractmethod
    def set_basic(
            self,
            name: str,

            msg: ta.Union[str, tuple, LoggingMsgFn],
            args: tuple,
    ) -> 'CaptureLoggingContext':
        raise NotImplementedError

    #

    class AlreadyCapturedError(Exception):
        pass

    class NotCapturedError(Exception):
        pass

    @abc.abstractmethod
    def capture(self) -> None:
        """Must be cooperatively called only from the expected locations."""

        raise NotImplementedError


@ta.final
class CaptureLoggingContextImpl(CaptureLoggingContext):
    @ta.final
    class NOT_SET:  # noqa
        def __new__(cls, *args, **kwargs):  # noqa
            raise TypeError

    #

    def __init__(
            self,
            level: LogLevel,
            *,
            time_ns: ta.Optional[int] = None,

            exc_info: LoggingExcInfoArg = False,

            caller: ta.Union[LoggingContextInfos.Caller, ta.Type[NOT_SET], None] = NOT_SET,
            stack_offset: int = 0,
            stack_info: bool = False,
    ) -> None:
        if time_ns is None:
            time_ns = time.time_ns()

        # Done early to not trample on sys.exc_info()
        exc = LoggingContextInfos.Exc.build(exc_info)

        self._infos: ta.Dict[ta.Type[LoggingContextInfo], LoggingContextInfo] = {}
        self._set_info(
            LoggingContextInfos.Level.build(level),
            exc,
            LoggingContextInfos.Time.build(time_ns),
        )

        if caller is not CaptureLoggingContextImpl.NOT_SET:
            self._infos[LoggingContextInfos.Caller] = caller
        else:
            self._stack_offset = stack_offset
            self._stack_info = stack_info

    def _set_info(self, *infos: ta.Optional[LoggingContextInfo]) -> 'CaptureLoggingContextImpl':
        for info in infos:
            if info is not None:
                self._infos[type(info)] = info
        return self

    def get_info(self, ty: ta.Type[LoggingContextInfoT]) -> ta.Optional[LoggingContextInfoT]:
        return self._infos.get(ty)

    ##

    def set_basic(
            self,
            name: str,

            msg: ta.Union[str, tuple, LoggingMsgFn],
            args: tuple,
    ) -> 'CaptureLoggingContextImpl':
        return self._set_info(
            LoggingContextInfos.Name(name),
            LoggingContextInfos.Msg.build(msg, *args),
        )

    ##

    _stack_offset: int
    _stack_info: bool

    def inc_stack_offset(self, ofs: int = 1) -> 'CaptureLoggingContext':
        if hasattr(self, '_stack_offset'):
            self._stack_offset += ofs
        return self

    _has_captured: bool = False

    def capture(self) -> None:
        if self._has_captured:
            raise CaptureLoggingContextImpl.AlreadyCapturedError
        self._has_captured = True

        if LoggingContextInfos.Caller not in self._infos:
            self._set_info(LoggingContextInfos.Caller.build(
                self._stack_offset + 1,
                stack_info=self._stack_info,
            ))

        if (caller := self[LoggingContextInfos.Caller]) is not None:
            self._set_info(LoggingContextInfos.SourceFile.build(
                caller.file_path,
            ))

        self._set_info(
            LoggingContextInfos.Thread.build(),
            LoggingContextInfos.Process.build(),
            LoggingContextInfos.Multiprocessing.build(),
            LoggingContextInfos.AsyncioTask.build(),
        )



import threading
import typing as ta

from ... import reflect as rfl
from ...funcs import match as mfs
from ..base.contexts import BaseContext
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import MarshalerFactory
from ..base.types import MarshalerMaker
from ..base.types import Unmarshaler
from ..base.types import UnmarshalerFactory
from ..base.types import UnmarshalerMaker


R = ta.TypeVar('R')
ContextT = ta.TypeVar('ContextT', bound=BaseContext)


##


class _InvalidatableFactory(mfs.MatchFn[[ContextT, rfl.Type], R]):
    def __init__(
            self,
            fn: ta.Callable[[], mfs.MatchFn[[ContextT, rfl.Type], R]],
            check_fn: ta.Callable[[], bool] | None = None,
    ) -> None:
        super().__init__()

        self._fn = fn
        self._check_fn: ta.Callable[[], bool] | None = check_fn

        self._lock = threading.RLock()

    #

    _f_: mfs.MatchFn[[ContextT, rfl.Type], R] | None = None

    def _invalidate(self) -> None:
        self._f_ = None

    def invalidate(self) -> None:
        with self._lock:
            self._invalidate()

    def _maybe_invalidate(self) -> None:
        if self._check_fn is not None:
            if self._check_fn():
                with self._lock:
                    if self._check_fn():
                        self._invalidate()

    def _f(self) -> mfs.MatchFn[[ContextT, rfl.Type], R]:
        self._maybe_invalidate()

        if (f := self._f_) is None:
            with self._lock:
                if (f := self._f_) is None:
                    f = self._f_ = self._fn()

        return f

    #

    def guard(self, ctx: ContextT, rty: rfl.Type) -> bool:
        return self._f().guard(ctx, rty)

    def fn(self, ctx: ContextT, rty: rfl.Type) -> R:
        return self._f()(ctx, rty)


##


class InvalidatableMarshalerFactory(MarshalerFactory):
    def __init__(
            self,
            fn: ta.Callable[[], MarshalerFactory],
            check_fn: ta.Callable[[], bool] | None = None,
    ) -> None:
        super().__init__()

        self._fn = fn
        self._u: _InvalidatableFactory[MarshalContext, Marshaler] = _InvalidatableFactory(
            lambda: fn().make_marshaler,
            check_fn,
        )

    def invalidate(self) -> None:
        self._u.invalidate()

    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self._u


class InvalidatableUnmarshalerFactory(UnmarshalerFactory):
    def __init__(
            self,
            fn: ta.Callable[[], UnmarshalerFactory],
            check_fn: ta.Callable[[], bool] | None = None,
    ) -> None:
        super().__init__()

        self._fn = fn
        self._u: _InvalidatableFactory[UnmarshalContext, Unmarshaler] = _InvalidatableFactory(
            lambda: fn().make_unmarshaler,
            check_fn,
        )

    def invalidate(self) -> None:
        self._u.invalidate()

    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self._u



import dataclasses as dc
import typing as ta

from ... import check
from ... import reflect as rfl
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..base.values import Value
from ..factories.simple import SimpleMarshalerFactory
from ..factories.simple import SimpleUnmarshalerFactory


##


@dc.dataclass(frozen=True)
class LiteralMarshaler(Marshaler):
    e: Marshaler
    vs: frozenset

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        return self.e.marshal(ctx, check.in_(o, self.vs))


class LiteralMarshalerFactory(SimpleMarshalerFactory):
    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.Literal)

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        lty = check.isinstance(rty, rfl.Literal)
        ety = check.single(set(map(type, lty.args)))
        return LiteralMarshaler(ctx.make(ety), frozenset(lty.args))


@dc.dataclass(frozen=True)
class LiteralUnmarshaler(Unmarshaler):
    e: Unmarshaler
    vs: frozenset

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        return check.in_(self.e.unmarshal(ctx, v), self.vs)


class LiteralUnmarshalerFactory(SimpleUnmarshalerFactory):
    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.Literal)

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        lty = check.isinstance(rty, rfl.Literal)
        ety = check.single(set(map(type, lty.args)))
        return LiteralUnmarshaler(ctx.make(ety), frozenset(lty.args))



import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import reflect as rfl
from ...funcs import match as mfs
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import MarshalerFactory
from ..base.types import MarshalerMaker
from ..base.types import Unmarshaler
from ..base.types import UnmarshalerFactory
from ..base.types import UnmarshalerMaker


R = ta.TypeVar('R')
C = ta.TypeVar('C')


##


@dc.dataclass(frozen=True)
class _TypeMapFactory(mfs.MatchFn[[C, rfl.Type], R]):
    m: ta.Mapping[rfl.Type, R] = dc.field(default_factory=dict)

    def __post_init__(self) -> None:
        for k in self.m:
            if not isinstance(k, rfl.TYPES):
                raise TypeError(k)

    def guard(self, ctx: C, rty: rfl.Type) -> bool:
        check.isinstance(rty, rfl.TYPES)
        return rty in self.m

    def fn(self, ctx: C, rty: rfl.Type) -> R:
        check.isinstance(rty, rfl.TYPES)
        try:
            return self.m[rty]
        except KeyError:
            raise mfs.MatchGuardError(ctx, rty)  # noqa


##


@dc.dataclass(frozen=True)
class TypeMapMarshalerFactory(
    _TypeMapFactory[MarshalContext, Marshaler],
    MarshalerFactory,
):
    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self  # noqa


@dc.dataclass(frozen=True)
class TypeMapUnmarshalerFactory(
    _TypeMapFactory[UnmarshalContext, Unmarshaler],
    UnmarshalerFactory,
):
    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self  # noqa



"""
TODO:
 - required_keys
"""
import abc
import typing as ta

from ... import dataclasses as dc
from ... import lang
from ..injector import AsyncInjector
from ..inspect import KwargsTarget
from ..providers import AsyncFnProvider
from ..providers import ConstProvider
from ..providers import CtorProvider
from ..providers import FnProvider
from ..providers import LinkProvider
from ..providers import Provider


##


class ProviderImpl(lang.Abstract):
    @property
    @abc.abstractmethod
    def providers(self) -> ta.Iterable[Provider]:
        raise NotImplementedError

    @abc.abstractmethod
    def provide(self, injector: AsyncInjector) -> ta.Awaitable[ta.Any]:
        raise NotImplementedError


##


@dc.dataclass(frozen=True, eq=False)
class InternalProvider(Provider):
    impl: ProviderImpl


##


@dc.dataclass(frozen=True, eq=False)
class AsyncCallableProviderImpl(ProviderImpl, lang.Final):
    p: AsyncFnProvider
    kt: KwargsTarget

    @property
    def providers(self) -> ta.Iterable[Provider]:
        return (self.p,)

    async def provide(self, injector: AsyncInjector) -> ta.Any:
        return await (await injector.inject(self.kt))


##


@dc.dataclass(frozen=True, eq=False)
class CallableProviderImpl(ProviderImpl, lang.Final):
    p: FnProvider | CtorProvider
    kt: KwargsTarget

    @property
    def providers(self) -> ta.Iterable[Provider]:
        return (self.p,)

    async def provide(self, injector: AsyncInjector) -> ta.Any:
        return await injector.inject(self.kt)


##


@dc.dataclass(frozen=True, eq=False)
class ConstProviderImpl(ProviderImpl, lang.Final):
    p: ConstProvider

    @property
    def providers(self) -> ta.Iterable[Provider]:
        return (self.p,)

    async def provide(self, injector: AsyncInjector) -> ta.Any:
        return self.p.v


##


@dc.dataclass(frozen=True, eq=False)
class LinkProviderImpl(ProviderImpl, lang.Final):
    p: LinkProvider

    @property
    def providers(self) -> ta.Iterable[Provider]:
        return (self.p,)

    async def provide(self, injector: AsyncInjector) -> ta.Any:
        return await injector.provide(self.p.k)



import abc
import typing as ta
import urllib.parse

from .. import check
from .. import lang
from ..logs import all as logs
from . import consts


log = logs.get_module_logger(globals())


##


Scope: ta.TypeAlias = ta.Mapping[str, ta.Any]
Message: ta.TypeAlias = ta.Mapping[str, ta.Any]
Recv: ta.TypeAlias = ta.Callable[[], ta.Awaitable[Message]]
Send: ta.TypeAlias = ta.Callable[[Message], ta.Awaitable[None]]
App: ta.TypeAlias = ta.Callable[[Scope, Recv, Send], ta.Awaitable[None]]
Wrapper: ta.TypeAlias = ta.Callable[[App, Scope, Recv, Send], ta.Awaitable[None]]


class App_(lang.Abstract):  # noqa
    @abc.abstractmethod
    async def __call__(self, scope: Scope, recv: Recv, send: Send) -> None:
        raise NotImplementedError


##


async def stub_lifespan(scope: Scope, recv: Recv, send: Send, *, verbose: bool = False) -> None:
    while True:
        message = await recv()
        if message['type'] == 'lifespan.startup':
            if verbose:
                log.info('Lifespan starting up')
            await send({'type': 'lifespan.startup.complete'})

        elif message['type'] == 'lifespan.shutdown':
            if verbose:
                log.info('Lifespan shutting down')
            await send({'type': 'lifespan.shutdown.complete'})
            return


##


async def start_response(
        send: Send,
        status: int,
        content_type: bytes = consts.CONTENT_TYPE_TEXT_UTF8,
        headers: ta.Sequence[tuple[bytes, bytes]] | None = None,
) -> None:
    await send({
        'type': 'http.response.start',
        'status': status,
        'headers': [
            (b'content-type', content_type),
            *(headers or ()),
        ],
    })


async def finish_response(
        send: Send,
        body: bytes = b'',
) -> None:
    await send({
        'type': 'http.response.body',
        'body': body,
    })


async def send_response(
        send: Send,
        status: int,
        content_type: bytes = consts.CONTENT_TYPE_TEXT_UTF8,
        headers: ta.Sequence[tuple[bytes, bytes]] | None = None,
        body: bytes = b'',
) -> None:
    await start_response(
        send,
        status=status,
        content_type=content_type, headers=headers,
    )
    await finish_response(
        send,
        body=body,
    )


async def redirect_response(
        send: Send,
        url: str,
        headers: ta.Sequence[tuple[bytes, bytes]] | None = None,
) -> None:
    log.info('Redirecting to %s', url)
    await send({
        'type': 'http.response.start',
        'status': 302,
        'headers': [
            (b'content-type', consts.CONTENT_TYPE_TEXT_UTF8),
            (b'location', url.encode()),
            *(headers or ()),
        ],
    })
    await send({
        'type': 'http.response.body',
        'body': b'',
    })


##


async def read_body(recv: Recv) -> bytes:
    body = b''
    more_body = True
    while more_body:
        message = await recv()
        body += message.get('body', b'')
        more_body = message.get('more_body', False)
    return body


async def read_form_body(recv: Recv) -> dict[bytes, bytes]:
    body = await read_body(recv)
    dct = urllib.parse.parse_qs(body)  # noqa
    return {k: check.single(v) for k, v in dct.items()}



from .helpers import (  # noqa
    assert_raises_star,
)

from . import marks  # noqa
from . import skip  # noqa

# Imported for convenience in things that import this but not lang.
from ...lang import breakpoint_on_exception  # noqa



import abc
import typing as ta

from .. import lang
from .elements import Elemental
from .elements import as_elements
from .inspect import KwargsTarget
from .keys import Key


if ta.TYPE_CHECKING:
    from .impl import sync as _sync
else:
    _sync = lang.proxy_import('.impl.sync', __package__)


T = ta.TypeVar('T')


##


class Injector(lang.Abstract):
    @abc.abstractmethod
    def try_provide(self, key: ta.Any) -> lang.Maybe[ta.Any]:
        raise NotImplementedError

    @abc.abstractmethod
    def provide(self, key: ta.Any) -> ta.Any:
        raise NotImplementedError

    @abc.abstractmethod
    def provide_kwargs(self, kt: KwargsTarget) -> ta.Mapping[str, ta.Any]:
        raise NotImplementedError

    @abc.abstractmethod
    def inject(self, obj: ta.Any) -> ta.Any:
        raise NotImplementedError

    def __getitem__(
            self,
            target: Key[T] | type[T],
    ) -> T:
        return self.provide(target)


def create_injector(*args: Elemental) -> Injector:
    return _sync.create_injector(as_elements(*args))



from ... import lang
from ..base.types import MarshalerFactory
from ..base.types import MarshalerMaker
from ..base.types import UnmarshalerFactory
from ..base.types import UnmarshalerMaker


##


class SimpleMarshalerFactory(
    MarshalerFactory,
    MarshalerMaker,
    lang.Abstract,
):
    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self  # noqa


class SimpleUnmarshalerFactory(
    UnmarshalerFactory,
    UnmarshalerMaker,
    lang.Abstract,
):
    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self  # noqa



from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from .clients.base import (  # noqa
        DEFAULT_ENCODING,

        is_success_status,

        HttpRequest,

        BaseHttpResponse,
        HttpResponse,

        HttpClientError,
        HttpStatusError,
    )

    from .clients.default import (  # noqa
        client,

        request,
    )

    from .clients.httpx import (  # noqa
        HttpxHttpClient,
    )

    from .clients.sync import (  # noqa
        StreamHttpResponse,

        close_response,
        closing_response,
        read_response,

        HttpClient,
    )

    from .clients.urllib import (  # noqa
        UrllibHttpClient,
    )

    from . import consts  # noqa

    from .cookies import (  # noqa
        CookieTooBigError,
        dump_cookie,
        parse_cookie,
    )

    from .dates import (  # noqa
        http_date,
        parse_date,
    )

    from .encodings import (  # noqa
        latin1_decode,
        latin1_encode,
    )

    from .headers import (  # noqa
        CanHttpHeaders,
        DuplicateHttpHeaderError,
        HttpHeaders,
        headers,
    )

    from .json import (  # noqa
        JSON_TAGGER,
        JsonTag,
        JsonTagger,
        json_dumps,
        json_loads,
    )

    from .multipart import (  # noqa
        MultipartData,
        MultipartEncoder,
        MultipartField,
    )



# ruff: noqa: UP006 UP007 UP043 UP045 UP046 UP047
# @omlish-lite
"""
A system for writing a python function once which can then be effectively used in both sync and async contexts -
including IO, under any (or no) event loop.

Where an 'async fn' returns an 'awaitable', a 'maysync fn' returns a 'maywaitable', which may be `await`'ed in async or
maysync contexts, or passed to `run_maysync` to be run sync.

For example, a maysync function `m_inc_int(x: int) -> int` would be used as such:

 - `run_maysync(m_inc_int(5)) == 6` in sync contexts
 - `assert await m_inc_int(5) == 6` in async and maysync contexts

Maysync fns may be constructed in two ways: either using `make_maysync`, providing an equivalent pair of sync and async
functions, or writing an 'maysync flavored' async python function. 'Maysync flavored' async functions are ones which
only call (await on) other maysync functions - the maysync machinery will ultimately call the appropriate 'leaf' sync or
async functions. Being regular python functions they are free to call whatever they like - for example doing sync IO -
but the point is to, ideally, route all IO through maysync functions such that the maysync code is fully efficiently
usable in any context.

Internally, it's not really correct to say that there is 'no event loop' in the maysync context - rather, each
'entrypoint' call to a maysync fn runs within its own tiny event loop.

===

TODO:
 - __del__
 - (test) maysync context managers
 - CancelledError
 - for debug, mask any running eventloop while running maysync code
 - `[CO_ASYNC_GENERATOR] = {k for k, v in dis.COMPILER_FLAG_NAMES.items() if v == 'ASYNC_GENERATOR'}` ? inspect is big..
  - works down to 3.8
 - make_maysync_from_sync can run with asyncio.run_in_thread
 - make_maysync_from_async can run with asyncio.run_soon

TODO OVERHAUL:
 - no more sync/async context, just one Context, and it means sync
 - make FpMaywaitable *not reusable*
  - `cannot reuse already awaited coroutine`
"""
import abc
import inspect
import sys
import threading
import typing as ta

from .abstract import Abstract


T = ta.TypeVar('T')
O = ta.TypeVar('O')
I = ta.TypeVar('I')

_MaysyncX = ta.TypeVar('_MaysyncX')
_MaysyncRS = ta.TypeVar('_MaysyncRS')
_MaysyncRA = ta.TypeVar('_MaysyncRA')


##


class AnyMaysyncFn(Abstract, ta.Generic[_MaysyncRS, _MaysyncRA]):
    def __init__(
            self,
            s: ta.Callable[..., _MaysyncRS],
            a: ta.Callable[..., _MaysyncRA],
    ) -> None:
        if s is None:
            raise TypeError(s)
        if a is None:
            raise TypeError(a)
        self._s, self._a = s, a

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._s!r}, {self._a!r})'

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        if hasattr(inspect, 'markcoroutinefunction'):
            inspect.markcoroutinefunction(cls.__call__)

    class FnPair(ta.NamedTuple):
        s: ta.Callable[..., ta.Any]
        a: ta.Callable[..., ta.Any]

    def fn_pair(self) -> ta.Optional[FnPair]:
        return AnyMaysyncFn.FnPair(
            self._s,
            self._a,
        )

    def __get__(self, instance, owner=None):
        return self.__class__(
            self._s.__get__(instance, owner),  # noqa
            self._a.__get__(instance, owner),  # noqa
        )

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        raise NotImplementedError


@ta.final
class MaywaitableAlreadyConsumedError(Exception):
    pass


class AnyMaywaitable(Abstract, ta.Generic[_MaysyncX]):
    @ta.final
    def __init__(
            self,
            x: _MaysyncX,
            args: ta.Tuple[ta.Any, ...],
            kwargs: ta.Mapping[str, ta.Any],
    ) -> None:
        self._x, self._args, self._kwargs = x, args, kwargs

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._x!r})'


##


@ta.final
class MaysyncFn(AnyMaysyncFn[T, ta.Awaitable[T]], ta.Generic[T]):  # noqa
    def __call__(self, *args, **kwargs):
        return Maywaitable(self, args, kwargs)


@ta.final
class Maywaitable(AnyMaywaitable[MaysyncFn[T]], ta.Awaitable[T]):
    _d: bool = False

    def __run_maysync__(self) -> T:
        if self._d or self._aw is not None:
            raise MaywaitableAlreadyConsumedError(self)

        self._d = True
        return self._x._s(*self._args, **self._kwargs)  # noqa

    #

    _aw: ta.Optional[ta.Awaitable[T]] = None

    def __await__(self) -> ta.Generator[ta.Any, None, T]:
        if (aw := self._aw) is None:
            if self._d:
                raise MaywaitableAlreadyConsumedError(self)

            if _MaysyncThreadLocal.run_context is None:
                aw = self._aw = self._x._a(*self._args, **self._kwargs)  # noqa
            else:
                aw = self._aw = _MaysyncFuture(self)

        return aw.__await__()


##


@ta.final
class MaysyncGeneratorFn(AnyMaysyncFn, ta.Generic[O, I]):  # noqa
    def __call__(self, *args, **kwargs):
        return MaysyncGenerator(self, args, kwargs)


@ta.final
class MaysyncGenerator(AnyMaywaitable[MaysyncGeneratorFn[O, I]], ta.AsyncGenerator[O, I]):
    def _make_sg(self) -> ta.Generator[O, I, None]:
        return self._x._s(*self._args, **self._kwargs)  # noqa

    _sg: ta.Optional[ta.Generator[O, I, None]] = None

    def _get_sg(self) -> ta.Generator[O, I, None]:
        if (sg := self._sg) is None:
            if self._ag is not None:
                raise MaywaitableAlreadyConsumedError(self)
            sg = self._sg = self._make_sg()
        return sg

    def __run_maysync__(self) -> ta.Generator[O, I, None]:
        return (self._sg if self._sg is not None else self._get_sg())

    #

    def _make_ag(self) -> ta.AsyncGenerator[O, I]:
        if _MaysyncThreadLocal.run_context is None:
            return self._x._a(*self._args, **self._kwargs)  # noqa

        async def inner():
            g = self._x._s(*self._args, **self._kwargs)  # noqa

            i: ta.Any = None
            e: ta.Any = None

            while True:
                try:
                    if e is not None:
                        o = g.throw(e)
                    else:
                        o = g.send(i)
                except StopIteration as ex:
                    if ex.value is not None:
                        raise TypeError(ex) from None
                    return

                i = None
                e = None

                try:
                    i = yield o
                except StopIteration as ex:  # noqa
                    raise NotImplementedError  # noqa
                except StopAsyncIteration as ex:  # noqa
                    raise NotImplementedError  # noqa
                except BaseException as ex:  # noqa
                    e = ex

        return inner()

    _ag: ta.Optional[ta.AsyncGenerator[O, I]] = None

    def _get_ag(self) -> ta.AsyncGenerator[O, I]:
        if (ag := self._ag) is None:
            if self._sg is not None:
                raise MaywaitableAlreadyConsumedError(self)
            ag = self._ag = self._make_ag()
        return ag

    def __anext__(self):
        return (self._ag if self._ag is not None else self._get_ag()).__anext__()

    def asend(self, value: I):
        return (self._ag if self._ag is not None else self._get_ag()).asend(value)

    def athrow(self, typ, val=None, tb=None):
        return (self._ag if self._ag is not None else self._get_ag()).athrow(typ, val, tb)

    def aclose(self):
        return (self._ag if self._ag is not None else self._get_ag()).aclose()


##


def make_maysync_fn(
        s: ta.Callable[..., T],
        a: ta.Callable[..., ta.Awaitable[T]],
) -> ta.Callable[..., ta.Awaitable[T]]:
    """Constructs a MaysyncFn from a (sync, async) function pair."""

    return MaysyncFn(s, a)


def make_maysync_generator_fn(
        s: ta.Callable[..., ta.Generator[O, I, None]],
        a: ta.Callable[..., ta.AsyncGenerator[O, I]],
) -> ta.Callable[..., ta.AsyncGenerator[O, I]]:
    """Constructs a MaysyncGeneratorFn from a (sync, async) generator function pair."""

    return MaysyncGeneratorFn(s, a)


@ta.overload
def make_maysync(
        s: ta.Callable[..., T],
        a: ta.Callable[..., ta.Awaitable[T]],
) -> ta.Callable[..., ta.Awaitable[T]]:
    ...


@ta.overload
def make_maysync(
        s: ta.Callable[..., ta.Generator[O, I, None]],
        a: ta.Callable[..., ta.AsyncGenerator[O, I]],
) -> ta.Callable[..., ta.AsyncGenerator[O, I]]:
    ...


def make_maysync(s, a):
    """
    Constructs a MaysyncFn or MaysyncGeneratorFn from a (sync, async) function pair, using `inspect.isasyncgenfunction`
    to determine the type.
    """

    if inspect.isasyncgenfunction(a):
        return MaysyncGeneratorFn(s, a)
    else:
        return MaysyncFn(s, a)


##


@ta.final
class _MaysyncThreadLocal(threading.local):
    run_context: ta.Optional['_MaysyncRunContext'] = None


@ta.final
class _MaysyncRunContext:
    def __init__(self, root: ta.Any) -> None:
        self._root = root

    def run(self, fn: ta.Callable[..., T], *args: ta.Any, **kwargs: ta.Any) -> T:
        prev = _MaysyncThreadLocal.run_context
        _MaysyncThreadLocal.run_context = self

        ph: ta.Any = sys.get_asyncgen_hooks()
        if ph.firstiter is not None or ph.finalizer is not None:
            sys.set_asyncgen_hooks(firstiter=None, finalizer=None)
        else:
            ph = None

        try:
            return fn(*args, **kwargs)

        finally:
            if ph is not None:
                sys.set_asyncgen_hooks(*ph)

            if _MaysyncThreadLocal.run_context is not self:
                raise RuntimeError
            _MaysyncThreadLocal.run_context = prev


def is_running_maysync() -> bool:
    return _MaysyncThreadLocal.run_context is not None


##


@ta.final
class _MaysyncFutureNotAwaitedError(RuntimeError):
    pass


@ta.final
class _MaysyncFuture(ta.Generic[T]):
    def __init__(
            self,
            x: Maywaitable[T],
    ) -> None:
        self._x = x

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._x!r}, done={self.done!r})'

    done: bool = False
    result: T
    error: ta.Optional[BaseException] = None

    def __await__(self) -> ta.Generator['_MaysyncFuture', None, T]:
        if not self.done:
            yield self
        if not self.done:
            raise _MaysyncFutureNotAwaitedError
        if self.error is not None:
            raise self.error
        else:
            return self.result

    def s(self) -> None:
        if self.done:
            return

        try:
            self.result = self._x._x._s(*self._x._args, **self._x._kwargs)  # noqa
        except BaseException as ex:  # noqa
            self.error = ex
        self.done = True

    async def a(self) -> None:
        if self.done:
            return

        try:
            self.result = await self._x._x._a(*self._x._args, **self._x._kwargs)  # noqa
        except BaseException as ex:  # noqa
            self.error = ex
        self.done = True


##


@ta.final
class _MaysyncDriver:
    def __init__(
            self,
            ctx: _MaysyncRunContext,
            coro: ta.Any,
    ) -> None:
        self.ctx, self.coro = ctx, coro

    value: ta.Any

    def __iter__(self) -> ta.Generator['_MaysyncFuture', None, None]:
        try:
            a = self.coro.__await__()
            try:
                g = iter(a)
                try:
                    while True:
                        try:
                            f = self.ctx.run(g.send, None)
                        except StopIteration as ex:
                            self.value = ex.value
                            return

                        if not isinstance(f, _MaysyncFuture):
                            raise TypeError(f)

                        yield f
                        del f

                finally:
                    if g is not a:
                        self.ctx.run(g.close)

            finally:
                self.ctx.run(a.close)

        finally:
            self.ctx.run(self.coro.close)


##


def run_maysync_fn(
        m: ta.Awaitable[T],
) -> T:
    if isinstance(m, Maywaitable):
        return m.__run_maysync__()

    for f in (drv := _MaysyncDriver(_MaysyncRunContext(m), m)):
        f.s()
        del f

    return drv.value


def run_maysync_generator_fn(
        m: ta.AsyncGenerator[O, I],
) -> ta.Generator[O, I, None]:
    if isinstance(m, MaysyncGenerator):
        return m.__run_maysync__()

    def inner():
        ctx = _MaysyncRunContext(m)

        try:
            i: ta.Any = None
            e: ta.Any = None

            while True:
                if e is not None:
                    coro = m.athrow(e)
                else:
                    coro = m.asend(i)

                i = None
                e = None

                drv = _MaysyncDriver(ctx, coro)
                di = iter(drv)
                try:
                    while True:
                        try:
                            f = next(di)
                        except StopAsyncIteration:
                            return
                        except StopIteration:
                            break

                        f.s()
                        del f

                finally:
                    di.close()

                o = drv.value
                try:
                    i = yield o
                except BaseException as ex:  # noqa
                    e = ex

        finally:
            for f in _MaysyncDriver(ctx, m.aclose()):
                f.s()

    return inner()


@ta.overload
def run_maysync(
        m: ta.Awaitable[T],
) -> T:
    ...


@ta.overload
def run_maysync(
        m: ta.AsyncGenerator[O, I],
) -> ta.Generator[O, I, None]:
    ...


def run_maysync(m):
    if hasattr(m, '__await__'):
        return run_maysync_fn(m)
    elif hasattr(m, '__aiter__'):
        return run_maysync_generator_fn(m)
    else:
        raise TypeError(m)


##


@ta.final
class RunMaysyncContextManager(ta.Generic[T]):
    def __init__(self, acm: ta.AsyncContextManager[T]) -> None:
        self._acm = acm

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._acm!r})'

    def __enter__(self) -> T:
        return run_maysync(self._acm.__aenter__())

    def __exit__(self, exc_type, exc_val, exc_tb):
        return run_maysync(self._acm.__aexit__(exc_type, exc_val, exc_tb))


run_maysync_context_manager = RunMaysyncContextManager


##


_MAYSYNC_MARK_ATTR = '__maysync__'


def mark_maysync(o):
    if not callable(o) or isinstance(o, AnyMaysyncFn):
        raise TypeError(o)
    setattr(o, _MAYSYNC_MARK_ATTR, True)
    return o


def is_maysync(o: ta.Any) -> bool:
    return isinstance(o, AnyMaysyncFn) or getattr(o, _MAYSYNC_MARK_ATTR, False)


mark_maysync(AnyMaysyncFn)



import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .keys import Key


##


class _Missing(lang.NotInstantiable):
    pass


class Provider(lang.Abstract):
    pass


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class AsyncFnProvider(Provider):
    fn: ta.Any = dc.xfield(validate=callable)


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class FnProvider(Provider):
    fn: ta.Any = dc.xfield(validate=callable)


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class CtorProvider(Provider):
    ty: type = dc.xfield(coerce=check.of_isinstance(type))


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class ConstProvider(Provider):
    v: ta.Any


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class LinkProvider(Provider):
    k: Key = dc.xfield(coerce=check.of_isinstance(Key))



import inspect
import typing as ta

from ... import check
from ... import collections as col
from ... import lang
from ... import reflect as rfl
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.options import Option
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..factories.simple import SimpleMarshalerFactory
from ..factories.simple import SimpleUnmarshalerFactory
from .marshal import ObjectMarshaler
from .metadata import FieldInfo
from .metadata import FieldInfos
from .metadata import FieldMetadata
from .unmarshal import ObjectUnmarshaler


##


def _is_namedtuple(rty: rfl.Type) -> bool:
    return (
        isinstance(rty, type) and
        issubclass(rty, tuple) and
        ta.NamedTuple in rfl.get_orig_bases(rty)
    )


def get_namedtuple_field_infos(
        ty: type,
        opts: col.TypeMap[Option] = col.TypeMap(),
) -> FieldInfos:
    check.arg(_is_namedtuple(ty), ty)

    sig = inspect.signature(ty)

    ret: list[FieldInfo] = []
    for param in sig.parameters.values():
        ret.append(FieldInfo(
            name=param.name,
            type=param.annotation,
            metadata=FieldMetadata(),

            marshal_name=param.name,
            unmarshal_names=[param.name],
        ))

    return FieldInfos(ret)


##


class NamedtupleMarshalerFactory(SimpleMarshalerFactory):
    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return _is_namedtuple(rty)

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        check.state(_is_namedtuple(rty))
        ty = check.isinstance(rty, type)
        check.state(not lang.is_abstract_class(ty))

        fis = get_namedtuple_field_infos(ty, ctx.options)

        fields = [
            (fi, ctx.make(fi.type))
            for fi in fis
        ]

        return ObjectMarshaler(
            fields,
        )


##


class NamedtupleUnmarshalerFactory(SimpleUnmarshalerFactory):
    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return _is_namedtuple(rty)

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        check.state(_is_namedtuple(rty))
        ty = check.isinstance(rty, type)
        check.state(not lang.is_abstract_class(ty))

        fis = get_namedtuple_field_infos(ty, ctx.options)

        d: dict[str, tuple[FieldInfo, Unmarshaler]] = {}
        defaults: dict[str, ta.Any] = {}

        for fi in fis:
            tup = (fi, ctx.make(fi.type))

            for un in fi.unmarshal_names:
                if un in d:
                    raise KeyError(f'Duplicate fields for name {un!r}: {fi.name!r}, {d[un][0].name!r}')
                d[un] = tup

            if fi.options.default.present:
                defaults[fi.name] = fi.options.default.must()

        return ObjectUnmarshaler(
            ty,
            d,
            defaults=defaults,
        )



import ast
import dataclasses as dc
import inspect
import pprint
import textwrap
import types
import typing as ta

from .. import lang


if ta.TYPE_CHECKING:
    import executing
else:
    executing = lang.proxy_import('executing')


##


class ArgsRenderer:
    """
    TODO:
     - kwargs
     - recursion
     - whatever pytest looks like
     - make sure not leaking sensitive data
    """

    def __init__(
            self,
            origin: types.TracebackType | types.FrameType | None = None,
            back: int = 1,
    ) -> None:
        super().__init__()

        self._origin = origin

        self._frame: types.FrameType | None
        if isinstance(origin, types.TracebackType):
            self._frame = origin.tb_frame
        elif isinstance(origin, types.FrameType):
            self._frame = origin
        elif origin is None:
            frame = inspect.currentframe()
            for _ in range(back + 1):
                if frame is None:
                    break
                frame = frame.f_back
            self._frame = frame
        else:
            raise TypeError(origin)

    def _get_indented_text(
            self,
            src: 'executing.Source',
            node: ast.AST,
    ) -> str:
        result = src.asttokens().get_text(node)
        if '\n' in result:
            result = ' ' * node.first_token.start[1] + result  # type: ignore
            result = textwrap.dedent(result)
        result = result.strip()
        return result

    def _val_to_string(self, obj: ta.Any) -> str:
        s = pprint.pformat(obj)
        s = s.replace('\\n', '\n')
        return s

    def _is_literal_expr(self, s: str) -> bool:
        try:
            ast.literal_eval(s)
        except Exception:  # noqa
            return False
        return True

    @dc.dataclass(frozen=True)
    class RenderedArg:
        val: str
        expr: str
        is_literal_expr: bool

        def __str__(self) -> str:
            if self.is_literal_expr:
                return self.val
            else:
                return f'({self.expr} = {self.val})'

    def render_args(
            self,
            *vals: ta.Any,
    ) -> ta.Sequence[RenderedArg] | None:
        if self._frame is None:
            return None

        call_node = executing.Source.executing(self._frame).node
        if not isinstance(call_node, ast.Call):
            return None

        source = executing.Source.for_frame(self._frame)

        exprs = [
            self._get_indented_text(source, arg)  # noqa
            for arg in call_node.args
        ]
        if len(exprs) != len(vals):
            return None

        return [
            self.RenderedArg(
                val=self._val_to_string(val),
                expr=expr,
                is_literal_expr=self._is_literal_expr(expr),
            )
            for val, expr in zip(vals, exprs)
        ]

    @classmethod
    def smoketest(cls) -> bool:
        def bar(z):
            return z + 1

        def foo(x, y):
            return cls().render_args(x, y)

        r = foo(1, bar(2))
        if r is None:
            return False

        x, y = r
        return (
            x == cls.RenderedArg(val='1', expr='1', is_literal_expr=True) and
            y == cls.RenderedArg(val='3', expr='bar(2)', is_literal_expr=False)
        )



# @omlish-lite
# ruff: noqa: UP006 UP007 UP043 UP045
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
# 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017 Python Software Foundation; All Rights Reserved" are retained in Python
# alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
import email.parser
import http
import typing as ta

from ....lite.check import check
from ..io import CoroHttpIo
from .errors import CoroHttpClientErrors
from .headers import CoroHttpClientHeaders
from .status import CoroHttpClientStatusLine


##


class CoroHttpClientResponse:
    # See RFC 2616 sec 19.6 and RFC 1945 sec 6 for details.

    # The bytes from the socket object are iso-8859-1 strings. See RFC 2616 sec 2.2 which notes an exception for
    # MIME-encoded text following RFC 2047. The basic status line parsing only accepts iso-8859-1.

    def __init__(self, method: str) -> None:
        super().__init__()

        state = self._State()
        state.method = method
        self._state = state

    class _State:
        closed: bool = False

        # If the response includes a content-length header, we need to make sure that the client doesn't read more than
        # the specified number of bytes. If it does, it will block until the server times out and closes the connection.
        # This will happen if a read is done (without a size) whether self.fp is buffered or not. So, no read by clients
        # unless they know what they are doing.
        method: str

        # The HttpResponse object is returned via urllib. The clients of http and urllib expect different attributes for
        # the headers. headers is used here and supports urllib. msg is provided as a backwards compatibility layer for
        # http clients.
        headers: email.message.Message

        # From the Status-Line of the response
        version: int  # HTTP-Version
        status: int  # Status-Code
        reason: str  # Reason-Phrase

        chunked: bool  # Is "chunked" being used?
        chunk_left: ta.Optional[int]  # Bytes left to read in current chunk
        length: ta.Optional[int]  # Number of bytes left in response
        will_close: bool  # Conn will close at end of response

    #

    def _read_status(self) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], CoroHttpClientStatusLine]:
        try:
            return (yield from CoroHttpClientStatusLine.read())
        except CoroHttpClientErrors.BadStatusLineError:
            self._close_conn()
            raise

    def _begin(self) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], None]:
        state = self._state

        check.state(not hasattr(state, 'headers'))

        # Read until we get a non-100 response
        while True:
            version, status, reason = yield from self._read_status()
            if status != http.HTTPStatus.CONTINUE:
                break

            # Skip the header from the 100 response
            skipped_headers = yield from CoroHttpClientHeaders.read_headers()  # noqa

            del skipped_headers

        state.status = status
        state.reason = reason.strip()
        if version in ('HTTP/1.0', 'HTTP/0.9'):
            # Some servers might still return '0.9', treat it as 1.0 anyway
            state.version = 10
        elif version.startswith('HTTP/1.'):
            # Use HTTP/1.1 code for HTTP/1.x where x>=1
            state.version = 11
        else:
            raise CoroHttpClientErrors.UnknownProtocolError(version)

        state.headers = yield from CoroHttpClientHeaders.parse_headers()

        # Are we using the chunked-style of transfer encoding?
        tr_enc = state.headers.get('transfer-encoding')
        if tr_enc and tr_enc.lower() == 'chunked':
            state.chunked = True
            state.chunk_left = None
        else:
            state.chunked = False

        # Will the connection close at the end of the response?
        state.will_close = self._check_close()

        # Do we have a Content-Length?
        # NOTE: RFC 2616, S4.4, #3 says we ignore this if tr_enc is 'chunked'
        state.length = None
        length = state.headers.get('content-length')
        if length and not state.chunked:
            try:
                state.length = int(length)
            except ValueError:
                state.length = None
            else:
                if state.length < 0:  # Ignore nonsensical negative lengths
                    state.length = None
        else:
            state.length = None

        # Does the body have a fixed length? (of zero)
        if (
                status in (http.HTTPStatus.NO_CONTENT, http.HTTPStatus.NOT_MODIFIED) or
                100 <= status < 200 or  # 1xx codes
                state.method == 'HEAD'
        ):
            state.length = 0

        # If the connection remains open, and we aren't using chunked, and a content-length was not provided, then
        # assume that the connection WILL close.
        if (
                not state.will_close and
                not state.chunked and
                state.length is None
        ):
            state.will_close = True

    #

    def _check_close(self) -> bool:
        conn = self._state.headers.get('connection')
        if getattr(self._state, 'version', None) == 11:
            # An HTTP/1.1 proxy is assumed to stay open unless explicitly closed.
            if conn and 'close' in conn.lower():
                return True
            return False

        # Some HTTP/1.0 implementations have support for persistent connections, using rules different than HTTP/1.1.

        # For older HTTP, Keep-Alive indicates persistent connection.
        if self._state.headers.get('keep-alive'):
            return False

        # At least Akamai returns a 'Connection: Keep-Alive' header, which was supposed to be sent by the client.
        if conn and 'keep-alive' in conn.lower():
            return False

        # Proxy-Connection is a netscape hack.
        pconn = self._state.headers.get('proxy-connection')
        if pconn and 'keep-alive' in pconn.lower():
            return False

        # Otherwise, assume it will close
        return True

    def _close_conn(self) -> None:
        self._state.closed = True

    def close(self) -> None:
        if not self._state.closed:
            self._close_conn()

    def is_closed(self) -> bool:
        """True if the connection is closed."""

        # NOTE: it is possible that we will not ever call self.close(). This case occurs when will_close is TRUE, length
        #   is None, and we read up to the last byte, but NOT past it.
        #
        # IMPLIES: if will_close is FALSE, then self.close() will ALWAYS be called, meaning self.is_closed() is
        #   meaningful.
        return self._state.closed

    #

    def read(self, amt: ta.Optional[int] = None) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], bytes]:
        """Read and return the response body, or up to the next amt bytes."""

        if self._state.closed:
            return b''

        if self._state.method == 'HEAD':
            self._close_conn()
            return b''

        if self._state.chunked:
            return (yield from self._read_chunked(amt))

        if amt is not None and amt >= 0:
            if self._state.length is not None and amt > self._state.length:
                # Clip the read to the "end of response"
                amt = self._state.length

            s = check.isinstance((yield CoroHttpIo.ReadIo(amt)), bytes)

            if not s and amt:
                # Ideally, we would raise IncompleteRead if the content-length wasn't satisfied, but it might break
                # compatibility.
                self._close_conn()

            elif self._state.length is not None:
                self._state.length -= len(s)
                if not self._state.length:
                    self._close_conn()

            return s

        else:
            # Amount is not given (unbounded read) so we must check self.length
            if self._state.length is None:
                s = check.isinstance((yield CoroHttpIo.ReadIo(None)), bytes)

            else:
                try:
                    s = yield from self._safe_read(self._state.length)
                except CoroHttpClientErrors.IncompleteReadError:
                    self._close_conn()
                    raise

                self._state.length = 0

            self._close_conn()  # We read everything
            return s

    def _read_next_chunk_size(self) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], int]:
        # Read the next chunk size from the file
        line = check.isinstance((yield CoroHttpIo.ReadLineIo(CoroHttpIo.MAX_LINE + 1)), bytes)
        if len(line) > CoroHttpIo.MAX_LINE:
            raise CoroHttpClientErrors.LineTooLongError(CoroHttpClientErrors.LineTooLongError.LineType.CHUNK_SIZE)

        i = line.find(b';')
        if i >= 0:
            line = line[:i]  # Strip chunk-extensions

        try:
            return int(line, 16)
        except ValueError:
            # Close the connection as protocol synchronisation is probably lost
            self._close_conn()
            raise

    def _read_and_discard_trailer(self) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], None]:
        # Read and discard trailer up to the CRLF terminator
        # NOTE: we shouldn't have any trailers!
        while True:
            line = check.isinstance((yield CoroHttpIo.ReadLineIo(CoroHttpIo.MAX_LINE + 1)), bytes)
            if len(line) > CoroHttpIo.MAX_LINE:
                raise CoroHttpClientErrors.LineTooLongError(CoroHttpClientErrors.LineTooLongError.LineType.TRAILER)

            if not line:
                # A vanishingly small number of sites EOF without sending the trailer
                break

            if line in (b'\r\n', b'\n', b''):
                break

    def _get_chunk_left(self) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], ta.Optional[int]]:
        # Return self.chunk_left, reading a new chunk if necessary. chunk_left == 0: at the end of the current chunk,
        # need to close it chunk_left == None: No current chunk, should read next. This function returns non-zero or
        # None if the last chunk has been read.
        chunk_left = self._state.chunk_left
        if not chunk_left:  # Can be 0 or None
            if chunk_left is not None:
                # We are at the end of chunk, discard chunk end
                yield from self._safe_read(2)  # Toss the CRLF at the end of the chunk

            try:
                chunk_left = yield from self._read_next_chunk_size()
            except ValueError:
                raise CoroHttpClientErrors.IncompleteReadError(b'') from None

            if chunk_left == 0:
                # Last chunk: 1*('0') [ chunk-extension ] CRLF
                yield from self._read_and_discard_trailer()

                # We read everything; close the 'file'
                self._close_conn()

                chunk_left = None

            self._state.chunk_left = chunk_left

        return chunk_left

    def _read_chunked(
            self,
            amt: ta.Optional[int] = None,
    ) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], bytes]:
        check.state(hasattr(self._state, 'chunked'))
        if amt is not None and amt < 0:
            amt = None
        value = []
        try:
            while (chunk_left := (yield from self._get_chunk_left())) is not None:
                if amt is not None and amt <= chunk_left:
                    value.append((yield from self._safe_read(amt)))
                    self._state.chunk_left = chunk_left - amt
                    break

                value.append((yield from self._safe_read(chunk_left)))
                if amt is not None:
                    amt -= chunk_left
                self._state.chunk_left = 0

            return b''.join(value)

        except CoroHttpClientErrors.IncompleteReadError as exc:
            raise CoroHttpClientErrors.IncompleteReadError(b''.join(value)) from exc

    def _safe_read(self, amt: int) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], bytes]:
        """
        Read the number of bytes requested.

        This function should be used when <amt> bytes "should" be present for reading. If the bytes are truly not
        available (due to EOF), then the IncompleteRead exception can be used to detect the problem.
        """

        data = check.isinstance((yield CoroHttpIo.ReadIo(amt)), bytes)
        if len(data) < amt:
            raise CoroHttpClientErrors.IncompleteReadError(data, amt - len(data))
        return data

    def peek(self, n: int = -1) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], bytes]:
        # Having this enables IOBase.readline() to read more than one byte at a time
        if self._state.closed or self._state.method == 'HEAD':
            return b''

        if self._state.chunked:
            return (yield from self._peek_chunked(n))

        return check.isinstance((yield CoroHttpIo.PeekIo(n)), bytes)

    def _readline(
            self,
            size: ta.Optional[int] = -1,
    ) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], bytes]:
        if size is None:
            size = -1
        else:
            try:
                size_index = size.__index__
            except AttributeError:
                raise TypeError(f'{size!r} is not an integer') from None
            else:
                size = size_index()

        # For backwards compatibility, a (slowish) readline().
        def nreadahead():
            readahead = yield from self.peek(1)
            if not readahead:
                return 1
            n = (readahead.find(b'\n') + 1) or len(readahead)
            if size >= 0:
                n = min(n, size)
            return n

        res = bytearray()
        while size < 0 or len(res) < size:
            b = (yield from self.read((yield from nreadahead())))
            if not b:
                break
            res += b
            if res.endswith(b'\n'):
                break

        return bytes(res)

    def readline(self, limit: int = -1) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], bytes]:
        if self._state.closed or self._state.method == 'HEAD':
            return b''

        if self._state.chunked:
            # Fallback to IOBase readline which uses peek() and read()
            return (yield from self._readline(limit))

        if self._state.length is not None and (limit < 0 or limit > self._state.length):
            limit = self._state.length

        result = check.isinstance((yield CoroHttpIo.ReadLineIo(limit)), bytes)

        if not result and limit:
            self._close_conn()

        elif self._state.length is not None:
            self._state.length -= len(result)
            if not self._state.length:
                self._close_conn()

        return result

    def _peek_chunked(self, n: int) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], bytes]:
        # Strictly speaking, _get_chunk_left() may cause more than one read, but that is ok, since that is to satisfy
        # the chunked protocol.
        try:
            chunk_left = yield from self._get_chunk_left()
        except CoroHttpClientErrors.IncompleteReadError:
            return b''  # Peek doesn't worry about protocol

        if chunk_left is None:
            return b''  # Eof

        # Peek is allowed to return more than requested. Just request the entire chunk, and truncate what we get.
        return check.isinstance((yield CoroHttpIo.PeekIo(chunk_left)), bytes)[:chunk_left]



import typing as ta

from ... import lang
from ..elements import Elements
from ..injector import AsyncInjector
from ..inspect import KwargsTarget
from ..keys import Key
from ..maysync import MaysyncInjector
from ..sync import Injector
from .elements import ElementCollection
from .injector import AsyncInjectorImpl


##


class MaysyncInjectorImpl(MaysyncInjector, lang.Final):
    _ai: AsyncInjector

    def try_provide(self, key: ta.Any) -> lang.Maybe[ta.Any]:
        return lang.run_maysync(self._ai.try_provide(key))

    def provide(self, key: ta.Any) -> ta.Any:
        return lang.run_maysync(self._ai.provide(key))

    def provide_kwargs(self, kt: KwargsTarget) -> ta.Mapping[str, ta.Any]:
        return lang.run_maysync(self._ai.provide_kwargs(kt))

    def inject(self, obj: ta.Any) -> ta.Any:
        return lang.run_maysync(self._ai.inject(obj))


def create_maysync_injector(es: Elements) -> MaysyncInjector:
    si = MaysyncInjectorImpl()
    ai = AsyncInjectorImpl(
        ElementCollection(es),
        internal_consts={
            Key(MaysyncInjector): si,
            Key(Injector): si,
        },
    )
    si._ai = ai  # noqa
    lang.run_maysync(ai._init())  # noqa
    return si



from .. import lang as _lang  # noqa


with _lang.auto_proxy_init(globals()):
    ##

    from .daemon import (  # noqa
        Daemon,
    )

    from .launching import (  # noqa
        Launcher,
    )

    from .services import (  # noqa
        Service,

        ServiceTarget,
        ServiceTargetRunner,

        ServiceConfigTarget,
        ServiceConfigTargetRunner,
    )

    from .spawning import (  # noqa
        Spawning,
        Spawn,
        Spawner,
        InProcessSpawner,
        spawner_for,

        MultiprocessingSpawning,
        MultiprocessingSpawner,

        ForkSpawning,
        ForkSpawner,

        ThreadSpawning,
        ThreadSpawner,
    )

    from .targets import (  # noqa
        Target,
        TargetRunner,
        target_runner_for,

        FnTarget,
        FnTargetRunner,

        NameTarget,
        NameTargetRunner,

        ExecTarget,
        ExecTargetRunner,
    )

    from .waiting import (  # noqa
        Wait,
        Waiter,
        waiter_for,

        SequentialWait,
        SequentialWaiter,

        FnWait,
        FnWaiter,

        ConnectWait,
        ConnectWaiter,
    )



# fmt: off
# ruff: noqa: I001
from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from .bimap import (  # noqa
        BiMap,

        make_bi_map,
    )

    from .coerce import (  # noqa
        abs_set,
        abs_set_of,
        abs_set_of_or_none,
        abs_set_or_none,
        frozenset_,
        frozenset_of,
        frozenset_of_or_none,
        frozenset_or_none,
        map,  # noqa
        map_of,
        map_of_or_none,
        map_or_none,
        opt_abs_set,
        opt_abs_set_of,
        opt_frozenset,
        opt_frozenset_of,
        opt_map,
        opt_map_of,
        opt_seq,
        opt_seq_of,
        seq,
        seq_of,
        seq_of_or_none,
        seq_or_none,
    )

    from .frozen import (  # noqa
        Frozen,
        FrozenDict,
        FrozenList,
        frozendict,
        frozenlist,
    )

    from .hasheq import (  # noqa
        HashEq,
        HashEqMap,
        HashEq_,
        hash_eq,
    )

    from .identity import (  # noqa
        IdentityKeyDict,
        IdentitySet,
        IdentityWeakKeyDictionary,
        IdentityWeakSet,
    )

    from . import kv  # noqa

    from .mappings import (  # noqa
        MissingDict,
        TypeMap,
        DynamicTypeMap,
        guarded_map_update,
        multikey_dict,
    )

    from .multimaps import (  # noqa
        MultiMap,

        SequenceMultiMap,
        AbstractSetMultiMap,

        BiMultiMap,
        InverseBiMultiMap,

        SequenceBiMultiMap,
        AbstractSetBiMultiMap,

        TupleBiMultiMap,
        seq_bi_multi_map,

        FrozensetBiMultiMap,
        abs_set_bi_multi_map,
    )

    from .ordered import (  # noqa
        OrderedFrozenSet,
        OrderedSet,
    )

    from .persistent.persistent import (  # noqa
        PersistentMap,
        PersistentMapping,
    )

    from .persistent.treapmap import (  # noqa
        TreapDict,
        TreapMap,
        new_treap_dict,
        new_treap_map,
    )

    from .ranked import (  # noqa
        RankedSeq,
        RankedSetSeq,
    )

    from .sorted.skiplist import (  # noqa
        SkipList,
        SkipListDict,
    )

    from .sorted.sorted import (  # noqa
        SortedCollection,
        SortedItems,
        SortedIter,
        SortedListDict,
        SortedMapping,
        SortedMutableMapping,
    )

    from .trie import (  # noqa
        Trie,
    )

    from .unmodifiable import (  # noqa
        Unmodifiable,
        UnmodifiableMapping,
        UnmodifiableSequence,
        UnmodifiableSet,
    )

    from .utils import (  # noqa
        PartitionResult,
        all_equal,
        all_not_equal,
        indexes,
        key_cmp,
        make_map,
        make_map_by,
        multi_map,
        multi_map_by,
        partition,
        unique,
    )



# ruff: noqa: UP007 UP043 UP045
import contextlib
import functools
import sys
import typing as ta


T = ta.TypeVar('T')
ExitStackedT = ta.TypeVar('ExitStackedT', bound='ExitStacked')
AsyncExitStackedT = ta.TypeVar('AsyncExitStackedT', bound='AsyncExitStacked')


##


class ExitStacked:
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        for a in ('__enter__', '__exit__'):
            for b in cls.__bases__:
                if b is ExitStacked:
                    continue
                try:
                    fn = getattr(b, a)
                except AttributeError:
                    pass
                else:
                    if fn is not getattr(ExitStacked, a):
                        raise TypeError(f'ExitStacked subclass {cls} must not not override {a} via {b}')

    _exit_stack: ta.Optional[contextlib.ExitStack] = None

    @contextlib.contextmanager
    def _exit_stacked_init_wrapper(self) -> ta.Iterator[None]:
        """
        Overridable wrapper around __enter__ which deliberately does not have access to an _exit_stack yet. Intended for
        things like wrapping __enter__ in a lock.
        """

        yield

    @ta.final
    def __enter__(self: ExitStackedT) -> ExitStackedT:
        """
        Final because any contexts entered during this init must be exited if any exception is thrown, and user
        overriding would likely interfere with that. Override `_enter_contexts` for such init.
        """

        with self._exit_stacked_init_wrapper():
            if self._exit_stack is not None:
                raise RuntimeError
            es = self._exit_stack = contextlib.ExitStack()
            es.__enter__()
            try:
                self._enter_contexts()
            except Exception:  # noqa
                es.__exit__(*sys.exc_info())
                raise
            return self

    @ta.final
    def __exit__(self, exc_type, exc_val, exc_tb):
        if (es := self._exit_stack) is None:
            return None
        try:
            self._exit_contexts()
        except Exception:  # noqa
            es.__exit__(*sys.exc_info())
            raise
        return es.__exit__(exc_type, exc_val, exc_tb)

    def _enter_contexts(self) -> None:
        pass

    def _exit_contexts(self) -> None:
        pass

    def _enter_context(self, cm: ta.ContextManager[T]) -> T:
        if (es := self._exit_stack) is None:
            raise RuntimeError
        return es.enter_context(cm)


class AsyncExitStacked:
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        for a in ('__aenter__', '__aexit__'):
            for b in cls.__bases__:
                if b is AsyncExitStacked:
                    continue
                try:
                    fn = getattr(b, a)
                except AttributeError:
                    pass
                else:
                    if fn is not getattr(AsyncExitStacked, a):
                        raise TypeError(f'AsyncExitStacked subclass {cls} must not not override {a} via {b}')

    _exit_stack: ta.Optional[contextlib.AsyncExitStack] = None

    @contextlib.asynccontextmanager
    async def _async_exit_stacked_init_wrapper(self) -> ta.AsyncGenerator[None, None]:
        yield

    @ta.final
    async def __aenter__(self: AsyncExitStackedT) -> AsyncExitStackedT:
        async with self._async_exit_stacked_init_wrapper():
            if self._exit_stack is not None:
                raise RuntimeError
            es = self._exit_stack = contextlib.AsyncExitStack()
            await es.__aenter__()
            try:
                await self._async_enter_contexts()
            except Exception:  # noqa
                await es.__aexit__(*sys.exc_info())
                raise
            return self

    @ta.final
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if (es := self._exit_stack) is None:
            return None
        try:
            await self._async_exit_contexts()
        except Exception:  # noqa
            await es.__aexit__(*sys.exc_info())
            raise
        return await es.__aexit__(exc_type, exc_val, exc_tb)

    async def _async_enter_contexts(self) -> None:
        pass

    async def _async_exit_contexts(self) -> None:
        pass

    def _enter_context(self, cm: ta.ContextManager[T]) -> T:
        if (es := self._exit_stack) is None:
            raise RuntimeError
        return es.enter_context(cm)

    async def _enter_async_context(self, cm: ta.AsyncContextManager[T]) -> T:
        if (es := self._exit_stack) is None:
            raise RuntimeError
        return await es.enter_async_context(cm)


##


@contextlib.contextmanager
def defer(fn: ta.Callable, *args: ta.Any, **kwargs: ta.Any) -> ta.Generator[ta.Callable, None, None]:
    if args or kwargs:
        fn = functools.partial(fn, *args, **kwargs)
    try:
        yield fn
    finally:
        fn()


@contextlib.asynccontextmanager
async def adefer(fn: ta.Awaitable) -> ta.AsyncGenerator[ta.Awaitable, None]:
    try:
        yield fn
    finally:
        await fn


##


@contextlib.contextmanager
def attr_setting(obj, attr, val, *, default=None):  # noqa
    not_set = object()
    orig = getattr(obj, attr, not_set)
    try:
        setattr(obj, attr, val)
        if orig is not not_set:
            yield orig
        else:
            yield default
    finally:
        if orig is not_set:
            delattr(obj, attr)
        else:
            setattr(obj, attr, orig)


##


class AsyncClosingManager(contextlib.AbstractAsyncContextManager):
    def __init__(self, thing):
        self.thing = thing

    async def __aenter__(self):
        return self.thing

    async def __aexit__(self, *exc_info):
        await self.thing.aclose()


aclosing = AsyncClosingManager



import dataclasses as dc
import typing as ta

from ... import check
from ... import lang
from ...lite.dataclasses import dataclass_cache_hash


##


_DEBUG = __debug__
# _DEBUG = True


@dc.dataclass(frozen=True)
class Value(lang.Abstract, lang.Sealed):
    meta: ta.Any | None = dc.field(default=None, kw_only=True)


#


@dc.dataclass(frozen=True)
class Scalar(Value, lang.Abstract):
    pass


@dataclass_cache_hash()
@dc.dataclass(frozen=True)
class Keyword(Scalar, lang.Final):
    s: str

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.s!r})'

    if _DEBUG:
        def __post_init__(self) -> None:
            check.isinstance(self.s, str)


@dataclass_cache_hash()
@dc.dataclass(frozen=True)
class Char(Scalar, lang.Final):
    c: str

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.c!r})'

    if _DEBUG:
        def __post_init__(self) -> None:
            check.isinstance(self.c, str)
            check.equal(len(self.c), 1)


@dataclass_cache_hash()
@dc.dataclass(frozen=True)
class Symbol(Scalar, lang.Final):
    n: str

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.n!r})'

    if _DEBUG:
        def __post_init__(self) -> None:
            check.non_empty_str(self.n)


#


@dc.dataclass(frozen=True)
class Collection(Value, lang.Abstract):
    pass


@dataclass_cache_hash()
@dc.dataclass(frozen=True)
class List(Collection, lang.Final):
    items: ta.Sequence[ta.Any]

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.items!r})'

    if _DEBUG:
        def __post_init__(self) -> None:
            check.isinstance(self.items, tuple)

    @classmethod
    def new(cls, items: ta.Iterable[ta.Any], *, meta: ta.Any | None = None) -> 'List':
        return cls(tuple(items), meta=meta)


@dataclass_cache_hash()
@dc.dataclass(frozen=True)
class Vector(Collection, lang.Final):
    items: ta.Sequence[ta.Any]

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.items!r})'

    if _DEBUG:
        def __post_init__(self) -> None:
            check.isinstance(self.items, tuple)

    @classmethod
    def new(cls, items: ta.Iterable[ta.Any], *, meta: ta.Any | None = None) -> 'Vector':
        return cls(tuple(items), meta=meta)


@dataclass_cache_hash()
@dc.dataclass(frozen=True)
class Set(Collection, lang.Final):
    items: ta.Sequence[ta.Any]

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.items!r})'

    if _DEBUG:
        def __post_init__(self) -> None:
            check.isinstance(self.items, tuple)

    @classmethod
    def new(cls, items: ta.Iterable[ta.Any], *, meta: ta.Any | None = None) -> 'Set':
        return cls(tuple(items), meta=meta)


@dataclass_cache_hash()
@dc.dataclass(frozen=True)
class Map(Collection, lang.Final):
    items: ta.Sequence[tuple[ta.Any, ta.Any]]

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.items!r})'

    if _DEBUG:
        def __post_init__(self) -> None:
            check.isinstance(self.items, tuple)
            for t in self.items:
                check.isinstance(t, tuple)
                check.equal(len(t), 2)

    @classmethod
    def new(cls, items: ta.Iterable[ta.Iterable[ta.Any]], *, meta: ta.Any | None = None) -> 'Map':
        return cls(tuple((k, v) for k, v in items), meta=meta)


#


@dataclass_cache_hash()
@dc.dataclass(frozen=True)
class Tagged(Value, lang.Final):
    t: str
    v: ta.Any

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.t!r}, {self.v!r})'

    if _DEBUG:
        def __post_init__(self) -> None:
            check.non_empty_str(self.t)



# ruff: noqa: UP006 UP045
# @omlish-lite
"""
https://www.brandur.org/logfmt
"""
import re
import typing as ta


##


_LOGFMT_ENCODE_WS_PAT = re.compile(r'[\s"]')


def logfmt_encode(
        dct: ta.Mapping[str, ta.Optional[ta.Any]],
        *,
        value_encoder: ta.Optional[ta.Callable[[ta.Any], ta.Optional[str]]] = None,
) -> str:
    def encode(s: str) -> str:
        if _LOGFMT_ENCODE_WS_PAT.search(s) is not None:
            s = s.replace('\\', '\\\\')
            s = s.replace('\n', '\\n')
            s = s.replace('"', '\\"')
            return f'"{s}"'
        else:
            return s

    if value_encoder is None:
        value_encoder = str

    lst: ta.List[str] = []
    for k, v in dct.items():
        ps: ta.List[str]
        if v is None:
            continue
        elif v is True:
            ps = [k]
        elif isinstance(v, str):
            ps = [k, v]
        elif (v := value_encoder(v)) is None:
            continue
        else:
            ps = [k, v]

        lst.append('='.join(map(encode, ps)))

    return ' '.join(lst)


#


_LOGFMT_DECODE_PAT_PART = r'[^"=\s]+|("([^\\]|(\\.))*?")'
_LOGFMT_DECODE_PAT = re.compile(
    rf'\s*'
    rf'(?P<key>{_LOGFMT_DECODE_PAT_PART})'
    rf'(=(?P<value>{_LOGFMT_DECODE_PAT_PART}))?'
    rf'\s*',
)


def logfmt_decode(
        line: str,
        *,
        value_decoder: ta.Optional[ta.Callable[[str], ta.Optional[ta.Any]]] = None,
) -> ta.Dict[str, ta.Any]:
    def unquote(s: str) -> str:
        if s.startswith('"'):
            if len(s) < 2 or not s.endswith('"'):
                raise ValueError(s)
            s = s[1:-1]
            s = s.replace('\\"', '"')
            s = s.replace('\\n', '\n')
            s = s.replace('\\\\', '\\')
        return s

    if value_decoder is None:
        def value_decoder(s: str) -> ta.Any:
            try:
                return int(s)
            except ValueError:
                pass
            try:
                return float(s)
            except ValueError:
                pass
            return s

    dct: ta.Dict[str, ta.Any] = {}
    p = 0
    while True:
        m = _LOGFMT_DECODE_PAT.match(line, p)
        if m is None:
            break
        if m.start() != p:
            raise RuntimeError

        k = unquote(m.group('key'))
        if k in dct:
            raise KeyError(k)

        vs = m.group('value')
        v: ta.Any
        if vs is None:
            v = True
        else:
            v = value_decoder(unquote(vs))

        dct[k] = v
        p = m.end()

    if p < len(line):
        raise ValueError(f'Unconsumed logfmt input: {line[p:]}')

    return dct



import signal
import typing as ta

import anyio.abc


##


async def install_shutdown_signal_handler(
        tg: anyio.abc.TaskGroup,
        event: anyio.Event | None = None,
        *,
        signals: ta.Iterable[int] = (signal.SIGINT, signal.SIGTERM),
        echo: bool = False,
) -> ta.Callable[..., ta.Awaitable[None]] | None:
    if event is None:
        event = anyio.Event()

    async def _handler(*, task_status=anyio.TASK_STATUS_IGNORED):
        with anyio.open_signal_receiver(*signals) as it:  # type: ignore
            task_status.started()
            async for signum in it:
                if echo:
                    if signum == signal.SIGINT:
                        print('Ctrl+C pressed!')
                    else:
                        print('Terminated!')
                event.set()
                return

    await tg.start(_handler)
    return event.wait



import dataclasses as dc
import enum
import typing as ta

from ... import check
from ... import reflect as rfl
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..base.values import Value
from ..factories.simple import SimpleMarshalerFactory
from ..factories.simple import SimpleUnmarshalerFactory


##


@dc.dataclass(frozen=True)
class EnumMarshaler(Marshaler):
    ty: type[enum.Enum]

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        return o.name


class EnumMarshalerFactory(SimpleMarshalerFactory):
    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, type) and issubclass(rty, enum.Enum)

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        ty = check.isinstance(rty, type)
        check.state(issubclass(ty, enum.Enum))
        return EnumMarshaler(ty)  # noqa


@dc.dataclass(frozen=True)
class EnumUnmarshaler(Unmarshaler):
    ty: type[enum.Enum]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return self.ty[check.isinstance(v, str)]


class EnumUnmarshalerFactory(SimpleUnmarshalerFactory):
    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, type) and issubclass(rty, enum.Enum)

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        ty = check.isinstance(rty, type)
        check.state(issubclass(ty, enum.Enum))
        return EnumUnmarshaler(ty)



import typing as ta

from .globals import OBJECT_SETATTR_GLOBAL
from .globals import FnGlobal
from .idents import SELF_DICT_IDENT
from .idents import SELF_IDENT


##


def build_attr_tuple_body_src_lines(
        obj_name: str,
        *attrs: str,
        prefix: str = '',
) -> list[str]:
    return [
        f'{prefix}{obj_name}.{a},'
        for a in attrs
    ]


def build_attr_kwargs_body_src_lines(
        obj_name: str,
        *attrs: str,
        prefix: str = '',
) -> list[str]:
    return [
        f'{prefix}{a}={obj_name}.{a},'
        for a in attrs
    ]


class SetattrSrcBuilder:
    def __init__(
            self,
            *,
            object_ident: str = SELF_IDENT,
            object_dict_ident: str | None = None,
    ) -> None:
        super().__init__()

        self._object_ident = object_ident
        self._object_dict_ident = object_dict_ident

        self._has_set_object_dict = False
        self._global_refs: set[FnGlobal] = set()

    @property
    def refs(self) -> ta.AbstractSet[FnGlobal]:
        return self._global_refs

    def __call__(
            self,
            name: str,
            value_src: str,
            *,
            frozen: bool,
            override: bool,
    ) -> list[str]:
        if override:
            if self._object_dict_ident is not None:
                return [f'{self._object_dict_ident}[{name!r}] = {value_src}']

            else:
                if not self._has_set_object_dict:
                    x = [
                        f'{SELF_DICT_IDENT} = {self._object_ident}.__dict__',
                    ]
                    self._has_set_object_dict = True
                else:
                    x = []
                return [
                    *x,
                    f'{SELF_DICT_IDENT}[{name!r}] = {value_src}',
                ]

        elif frozen:
            self._global_refs.add(OBJECT_SETATTR_GLOBAL)
            return [f'{OBJECT_SETATTR_GLOBAL.ident}({self._object_ident}, {name!r}, {value_src})']

        else:
            return [f'{self._object_ident}.{name} = {value_src}']



import dataclasses as dc
import typing as ta

from ... import reflect as rfl
from ...funcs import match as mfs
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.errors import ForbiddenTypeError
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..factories.simple import SimpleMarshalerFactory
from ..factories.simple import SimpleUnmarshalerFactory


C = ta.TypeVar('C')
R = ta.TypeVar('R')


##


@dc.dataclass(frozen=True)
class ForbiddenTypeFactory(mfs.MatchFn[[C, rfl.Type], R]):
    rtys: ta.AbstractSet[rfl.Type]

    def guard(self, ctx: C, rty: rfl.Type) -> bool:
        return rty in self.rtys

    def fn(self, ctx: C, rty: rfl.Type) -> R:
        raise ForbiddenTypeError(rty)


@dc.dataclass(frozen=True)
class ForbiddenTypeMarshalerFactory(
    ForbiddenTypeFactory[MarshalContext, Marshaler],
    SimpleMarshalerFactory,
):
    pass


@dc.dataclass(frozen=True)
class ForbiddenTypeUnmarshalerFactory(
    ForbiddenTypeFactory[UnmarshalContext, Unmarshaler],
    SimpleUnmarshalerFactory,
):
    pass



import bisect
import collections
import math
import typing as ta

from .. import cached
from .. import check


##


def get_quantile(sorted_data: ta.Sequence[float], q: float) -> float:
    q = float(q)
    check.arg(0.0 <= q <= 1.0)
    data, n = sorted_data, len(sorted_data)
    idx = q / 1.0 * (n - 1)
    idx_f, idx_c = math.floor(idx), math.ceil(idx)
    if idx_f == idx_c:
        return data[idx_f]
    return (data[idx_f] * (idx_c - idx)) + (data[idx_c] * (idx - idx_f))


##


class Stats(ta.Sequence[float]):
    """
    ~ https://github.com/mahmoud/boltons/blob/47e0c3bfcbd3291a1366f34069f23e43659717cd/boltons/statsutils.py
    """

    def __init__(
            self,
            data: ta.Sequence[float],
            *,
            default: float = 0.,
            eq: ta.Callable[[float, float], bool] = lambda a, b: a == b,
    ) -> None:
        super().__init__()

        self._kwargs: ta.Any = dict(
            default=default,
            eq=eq,
        )

        self._data = data
        self._default = default
        self._eq = eq

    @property
    def data(self) -> ta.Sequence[float]:
        return self._data

    @property
    def default(self) -> float:
        return self._default

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> ta.Iterator[float]:
        return iter(self.data)

    def __getitem__(self, index: ta.Any) -> float:  # type: ignore
        return self._data[index]

    @cached.property
    def sorted(self) -> ta.Sequence[float]:
        return sorted(self.data)

    @cached.property
    def mean(self) -> float:
        return sum(self.data, 0.0) / len(self.data)

    @cached.property
    def max(self) -> float:
        return max(self.data)

    @cached.property
    def min(self) -> float:
        return min(self.data)

    def get_quantile(self, q: float) -> float:
        if not self.data:
            return self.default
        return get_quantile(self.sorted, q)

    @cached.property
    def median(self) -> float:
        return self.get_quantile(0.5)

    def get_pow_diffs(self, power: float) -> list[float]:
        m = self.mean
        return [(v - m) ** power for v in self.data]

    @cached.property
    def variance(self) -> float:
        return Stats(self.get_pow_diffs(2)).mean

    @cached.property
    def std_dev(self) -> float:
        return self.variance ** 0.5

    @cached.property
    def median_abs_dev(self) -> float:
        x = self.median
        return Stats([abs(x - v) for v in self.sorted]).median

    @cached.property
    def rel_std_dev(self) -> float:
        abs_mean = abs(self.mean)
        if abs_mean:
            return self.std_dev / abs_mean
        else:
            return self.default

    @cached.property
    def skewness(self) -> float:
        data, s_dev = self.data, self.std_dev
        if len(data) > 1 and s_dev > 0:
            return (sum(self.get_pow_diffs(3)) / float((len(data) - 1) * (s_dev ** 3)))
        else:
            return self.default

    @cached.property
    def kurtosis(self) -> float:
        data, s_dev = self.data, self.std_dev
        if len(data) > 1 and s_dev > 0:
            return (sum(self.get_pow_diffs(4)) / float((len(data) - 1) * (s_dev ** 4)))
        else:
            return 0.0

    @cached.property
    def iqr(self) -> float:
        return self.get_quantile(0.75) - self.get_quantile(0.25)

    @cached.property
    def trimean(self) -> float:
        return (self.get_quantile(0.25) + (2 * self.get_quantile(0.5)) + self.get_quantile(0.75)) / 4.0

    def get_zscore(self, value: float) -> float:
        mean = self.mean
        if self._eq(self.std_dev, 0.0):
            if self._eq(value, mean):
                return 0
            if value > mean:
                return float('inf')
            if value < mean:
                return float('-inf')
        return (float(value) - mean) / self.std_dev

    def trim_relative(self, amount: float = 0.15) -> 'Stats':
        trim = float(amount)
        check.arg(0.0 <= trim < 0.5)
        size = len(self.data)
        size_diff = int(size * trim)
        if self._eq(size_diff, 0.0):
            return self
        return Stats(self.sorted[size_diff:-size_diff], **self._kwargs)

    def get_bin_bounds(
            self,
            count: int | None = None,
            with_max: bool = False,
    ) -> list[float]:
        if not self.data:
            return [0.0]

        data = self.data
        len_data, min_data, max_data = len(data), min(data), max(data)

        if len_data < 4:
            if not count:
                count = len_data
            dx = (max_data - min_data) / float(count)
            bins = [min_data + (dx * i) for i in range(count)]

        elif count is None:
            # freedman algorithm for fixed-width bin selection
            q25, q75 = self.get_quantile(0.25), self.get_quantile(0.75)
            dx = 2 * (q75 - q25) / (len_data ** (1 / 3.0))
            bin_count = max(1, math.ceil((max_data - min_data) / dx))
            bins = [min_data + (dx * i) for i in range(bin_count + 1)]
            bins = [b for b in bins if b < max_data]

        else:
            dx = (max_data - min_data) / float(count)
            bins = [min_data + (dx * i) for i in range(count)]

        if with_max:
            bins.append(float(max_data))

        return bins

    def get_histogram_counts(
            self,
            bins: list[float] | int | None = None,
            *,
            bin_digits: int = 1,
    ) -> list[tuple[float, int]]:
        bin_digits = int(bin_digits)
        if not bins:
            bins = self.get_bin_bounds()
        elif isinstance(bins, int):
            bins = self.get_bin_bounds(bins)
        else:
            bins = [float(x) for x in bins]
            if self.min < bins[0]:
                bins = [self.min, *bins]

        round_factor = 10.0 ** bin_digits
        bins = [math.floor(b * round_factor) / round_factor for b in bins]
        bins = sorted(set(bins))

        idxs = [bisect.bisect(bins, d) - 1 for d in self.data]
        count_map = collections.Counter(idxs)
        bin_counts = [(b, count_map.get(i, 0)) for i, b in enumerate(bins)]
        return bin_counts



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - optimization of just using ChainMap when override?
  - TypedLoggerBindings as Abstract, FullTLB and LimitedTLB subtypes, limited
 - TypedLoggerBindingsBuilder ?
"""
import abc
import collections
import functools
import itertools
import typing as ta

from ...lite.abstract import Abstract
from .types import TYPED_LOGGER_VALUE_OR_PROVIDER_OR_ABSENT_TYPES
from .types import TYPED_LOGGER_VALUE_OR_PROVIDER_TYPES
from .types import DefaultTypedLoggerValue
from .types import MultiTypedLoggerValue
from .types import TypedLoggerConstFieldValue
from .types import TypedLoggerField
from .types import TypedLoggerFieldValue
from .types import TypedLoggerValue
from .types import TypedLoggerValueOrAbsent
from .types import TypedLoggerValueOrProvider  # noqa
from .types import TypedLoggerValueOrProviderOrAbsent
from .types import TypedLoggerValueProvider


TypedLoggerBindingItem = ta.Union['TypedLoggerField', 'TypedLoggerValueOrProvider', 'TypedLoggerBindings', 'TypedLoggerValueWrapper']  # ta.TypeAlias  # noqa

ChainTypedLoggerBindingItem = ta.Union['TypedLoggerField', 'TypedLoggerValueOrProvider']  # ta.TypeAlias

TypedLoggerValueWrapperFn = ta.Callable[[ta.Any], 'TypedLoggerValue']  # ta.TypeAlias


##


class TypedLoggerDuplicateBindingsError(Exception):
    def __init__(
            self,
            *,
            keys: ta.Optional[ta.Dict[str, ta.List[TypedLoggerFieldValue]]] = None,
            values: ta.Optional[ta.Dict[ta.Type[TypedLoggerValue], ta.List[TypedLoggerValueOrProviderOrAbsent]]] = None,
            wrappers: ta.Optional[ta.Dict[type, ta.List[TypedLoggerValueWrapperFn]]] = None,
    ) -> None:
        super().__init__()

        self.keys = keys
        self.values = values
        self.wrappers = wrappers

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}(' +
            ', '.join([
                *([f'keys={self.keys!r}'] if self.keys is not None else []),
                *([f'values={self.values!r}'] if self.values is not None else []),
                *([f'wrappers={self.wrappers!r}'] if self.wrappers is not None else []),
            ]) +
            f')'
        )


class TypedLoggerBindings(Abstract):
    @property
    @abc.abstractmethod
    def key_map(self) -> ta.Mapping[str, TypedLoggerFieldValue]:
        raise NotImplementedError

    @abc.abstractmethod
    def lookup_value(self, cls: ta.Type[TypedLoggerValue]) -> ta.Optional[TypedLoggerValueOrProviderOrAbsent]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def const_value_map(self) -> ta.Mapping[ta.Type[TypedLoggerValue], TypedLoggerValueOrAbsent]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def value_wrapper_fn(self) -> ta.Optional[TypedLoggerValueWrapperFn]:
        raise NotImplementedError

    #

    @ta.final
    class _Visitor(ta.NamedTuple):
        accept_key: ta.Callable[[str, TypedLoggerFieldValue], None]
        accept_keys: ta.Callable[[ta.Iterable[ta.Tuple[str, TypedLoggerFieldValue]]], None]

        accept_value: ta.Callable[[ta.Type[TypedLoggerValue], TypedLoggerValueOrProviderOrAbsent], None]
        accept_values: ta.Callable[[ta.Iterable[ta.Tuple[ta.Type[TypedLoggerValue], TypedLoggerValueOrProviderOrAbsent]]], None]  # noqa

        accept_multi_value: ta.Callable[[ta.Type[MultiTypedLoggerValue], MultiTypedLoggerValue], None]  # noqa
        accept_multi_values: ta.Callable[[ta.Iterable[ta.Tuple[ta.Type[MultiTypedLoggerValue], MultiTypedLoggerValue]]], None]  # noqa

        accept_const_values: ta.Callable[[ta.Iterable[ta.Tuple[ta.Type[TypedLoggerValue], TypedLoggerValueOrAbsent]]], None]  # noqa

        accept_value_wrapping: ta.Callable[[ta.Union['TypedLoggerValueWrapper', 'FullTypedLoggerBindings._ValueWrappingState']], None]  # noqa

    @abc.abstractmethod
    def _typed_logger_visit_bindings(self, vst: _Visitor) -> None:
        raise NotImplementedError


##


@ta.final
class FullTypedLoggerBindings(TypedLoggerBindings):
    def __init__(
            self,
            *items: TypedLoggerBindingItem,
            override: bool = False,
            override_multis: bool = False,
    ) -> None:
        kd: ta.Dict[str, TypedLoggerFieldValue] = {}
        dup_kd: ta.Dict[str, ta.List[TypedLoggerFieldValue]] = {}

        if not override:
            def add_kd(kd_k: str, kd_v: TypedLoggerFieldValue) -> None:  # noqa
                if kd_k in kd:
                    dup_kd.setdefault(kd_k, []).append(kd_v)
                else:
                    kd[kd_k] = kd_v

            def add_kds(it: 'ta.Iterable[ta.Tuple[str, TypedLoggerFieldValue]]') -> None:  # noqa
                collections.deque(itertools.starmap(add_kd, it), maxlen=0)

        else:
            add_kd = kd.__setitem__  # type: ignore[assignment]
            add_kds = kd.update  # type: ignore[assignment]

        #

        vd: ta.Dict[ta.Type[TypedLoggerValue], TypedLoggerValueOrProviderOrAbsent] = {}
        dup_vd: ta.Dict[ta.Type[TypedLoggerValue], ta.List[TypedLoggerValueOrProviderOrAbsent]] = {}

        if not override:
            def add_vd(vd_k: 'ta.Type[TypedLoggerValue]', vd_v: TypedLoggerValueOrProviderOrAbsent) -> None:  # noqa
                if vd_k in vd:
                    dup_vd.setdefault(vd_k, []).append(vd_v)
                else:
                    vd[vd_k] = vd_v

            def add_vds(it: 'ta.Iterable[ta.Tuple[ta.Type[TypedLoggerValue], TypedLoggerValueOrProviderOrAbsent]]') -> None:  # noqa
                collections.deque(itertools.starmap(add_vd, it), maxlen=0)

        else:
            add_vd = vd.__setitem__  # type: ignore[assignment]
            add_vds = vd.update  # type: ignore[assignment]

        #

        mvd: ta.Dict[ta.Type[MultiTypedLoggerValue], ta.List[MultiTypedLoggerValue]] = {}

        if override_multis:
            def add_mvd(mvd_k: 'ta.Type[MultiTypedLoggerValue]', mvd_v: MultiTypedLoggerValue) -> None:  # noqa
                try:
                    mvd_l = mvd[mvd_k]
                except KeyError:
                    mvd_l = mvd[mvd_k] = []
                mvd_l.append(mvd_v)

            def add_mvds(it: 'ta.Iterable[ta.Tuple[ta.Type[MultiTypedLoggerValue], MultiTypedLoggerValue]]') -> None:
                collections.deque(itertools.starmap(add_mvd, it), maxlen=0)

        else:
            add_mvd = mvd.__setitem__  # type: ignore[assignment]
            add_mvds = mvd.update  # type: ignore[assignment]

        #

        cvd: ta.Dict[ta.Type[TypedLoggerValue], TypedLoggerValueOrAbsent] = {}

        vwl: ta.List[ta.Union[TypedLoggerValueWrapper, FullTypedLoggerBindings._ValueWrappingState]] = []

        vst = TypedLoggerBindings._Visitor(  # noqa
            add_kd,
            add_kds,

            add_vd,
            add_vds,

            add_mvd,
            add_mvds,

            cvd.update,

            vwl.append,
        )

        for o in items:
            o._typed_logger_visit_bindings(vst)  # noqa

        #

        dup_vwd: ta.Optional[ta.Dict[type, ta.List[TypedLoggerValueWrapperFn]]] = None

        vws: ta.Optional[FullTypedLoggerBindings._ValueWrappingState] = None
        vwf: ta.Optional[TypedLoggerValueWrapperFn] = None

        if vwl:
            if len(vwl) == 1 and isinstance((vwl0 := vwl[0]), FullTypedLoggerBindings._ValueWrappingState):
                vws = vwl0

            else:
                vwd: ta.Dict[type, TypedLoggerValueWrapperFn] = {}

                add_vwd: ta.Callable[[type, TypedLoggerValueWrapperFn], None]

                if not override:
                    dup_vwd = {}

                    def add_vwd(vw_ty: type, vw_fn: TypedLoggerValueWrapperFn) -> None:  # noqa
                        if vw_ty in vwd:
                            dup_vwd.setdefault(vw_ty, []).append(vw_fn)
                        else:
                            vwd[vw_ty] = vw_fn

                else:
                    add_vwd = vwd.__setitem__

                for vo in vwl:
                    vo._typed_logger_visit_value_wrappers(add_vwd)  # noqa

                if vwd and not dup_vwd:
                    vws = FullTypedLoggerBindings._ValueWrappingState(vwd)
                    vwf = vws.sdf

        #

        if dup_kd or dup_vd or dup_vwd:
            raise TypedLoggerDuplicateBindingsError(
                keys=dup_kd or None,
                values=dup_vd or None,
                wrappers=dup_vwd or None,
            )

        self._key_map: ta.Mapping[str, TypedLoggerFieldValue] = kd
        self._value_map: ta.Mapping[ta.Type[TypedLoggerValue], TypedLoggerValueOrProviderOrAbsent] = vd

        self._const_value_map: ta.Mapping[ta.Type[TypedLoggerValue], TypedLoggerValueOrAbsent] = cvd

        self._value_wrapping = vws
        self._value_wrapper_fn = vwf

    #

    @property
    def key_map(self) -> ta.Mapping[str, TypedLoggerFieldValue]:
        return self._key_map

    def lookup_value(self, cls: ta.Type[TypedLoggerValue]) -> ta.Optional[TypedLoggerValueOrProviderOrAbsent]:
        return self._value_map.get(cls)

    @property
    def const_value_map(self) -> ta.Mapping[ta.Type[TypedLoggerValue], TypedLoggerValueOrAbsent]:
        return self._const_value_map

    @property
    def value_wrapper_fn(self) -> ta.Optional[TypedLoggerValueWrapperFn]:
        return self._value_wrapper_fn

    #

    @ta.final
    class _ValueWrappingState:
        def __init__(self, dct: ta.Mapping[type, TypedLoggerValueWrapperFn]) -> None:
            self.dct = dct

            @functools.singledispatch
            def wrap_value(o: ta.Any) -> TypedLoggerValue:
                raise UnhandledTypedValueWrapperTypeError(o)

            collections.deque(itertools.starmap(wrap_value.register, self.dct.items()), maxlen=0)

            self.sdf = wrap_value

        #

        def _typed_logger_visit_value_wrappers(self, fn: ta.Callable[[type, TypedLoggerValueWrapperFn], None]) -> None:  # noqa
            collections.deque(itertools.starmap(fn, self.dct.items()), maxlen=0)

    #

    def _typed_logger_visit_bindings(self, vst: TypedLoggerBindings._Visitor) -> None:
        vst.accept_keys(self._key_map.items())
        vst.accept_values(self._value_map.items())

        vst.accept_const_values(self._const_value_map.items())

        if (vws := self._value_wrapping) is not None:
            vst.accept_value_wrapping(vws)


##


class ChainTypedLoggerBindingsUnhandledItemError(Exception):
    pass


@ta.final
class ChainTypedLoggerBindings(TypedLoggerBindings):
    def __init__(
            self,
            parent: TypedLoggerBindings,
            *items: ChainTypedLoggerBindingItem,
    ) -> None:
        kd: ta.Dict[str, TypedLoggerFieldValue] = {}
        dup_kd: ta.Dict[str, ta.List[TypedLoggerFieldValue]] = {}

        def add_kd(kd_k: str, kd_v: TypedLoggerFieldValue) -> None:  # noqa
            if kd_k in kd:
                dup_kd.setdefault(kd_k, []).append(kd_v)
            else:
                kd[kd_k] = kd_v

        def add_kds(it: 'ta.Iterable[ta.Tuple[str, TypedLoggerFieldValue]]') -> None:  # noqa
            collections.deque(itertools.starmap(add_kd, it), maxlen=0)

        #

        vd: ta.Dict[ta.Type[TypedLoggerValue], TypedLoggerValueOrProviderOrAbsent] = {}
        dup_vd: ta.Dict[ta.Type[TypedLoggerValue], ta.List[TypedLoggerValueOrProviderOrAbsent]] = {}

        pcv = dict(parent.const_value_map)

        def add_vd(vd_k: 'ta.Type[TypedLoggerValue]', vd_v: TypedLoggerValueOrProviderOrAbsent) -> None:  # noqa
            if vd_k in vd:
                dup_vd.setdefault(vd_k, []).append(vd_v)
            else:
                vd[vd_k] = vd_v
                pcv.pop(vd_k, None)

        def add_vds(it: 'ta.Iterable[ta.Tuple[ta.Type[TypedLoggerValue], TypedLoggerValueOrProviderOrAbsent]]') -> None:  # noqa
            collections.deque(itertools.starmap(add_vd, it), maxlen=0)

        #

        mvd: ta.Dict[ta.Type[MultiTypedLoggerValue], ta.List[MultiTypedLoggerValue]] = {}

        def add_mvd(mvd_k: 'ta.Type[MultiTypedLoggerValue]', mvd_v: MultiTypedLoggerValue) -> None:  # noqa
            try:
                mvd_l = mvd[mvd_k]
            except KeyError:
                mvd_l = mvd[mvd_k] = []
            mvd_l.append(mvd_v)

        def add_mvds(it: 'ta.Iterable[ta.Tuple[ta.Type[MultiTypedLoggerValue], MultiTypedLoggerValue]]') -> None:
            collections.deque(itertools.starmap(add_mvd, it), maxlen=0)

        #

        cvd: ta.Dict[ta.Type[TypedLoggerValue], TypedLoggerValueOrAbsent] = {}

        def bad_type(*args, **kwargs):  # noqa
            raise ChainTypedLoggerBindingsUnhandledItemError

        vst = TypedLoggerBindings._Visitor(  # noqa
            add_kd,
            add_kds,

            add_vd,
            add_vds,

            add_mvd,
            add_mvds,

            cvd.update,

            bad_type,
        )

        for o in items:
            try:
                o._typed_logger_visit_bindings(vst)  # noqa
            except ChainTypedLoggerBindingsUnhandledItemError:
                raise ChainTypedLoggerBindingsUnhandledItemError(o) from None

        #

        if dup_kd or dup_vd:
            raise TypedLoggerDuplicateBindingsError(
                keys=dup_kd or None,
                values=dup_vd or None,
            )

        self._parent = parent

        self._key_map = {**parent.key_map, **kd}
        self._self_value_map = vd
        self._const_value_map = {**pcv, **cvd}

    @property
    def key_map(self) -> ta.Mapping[str, TypedLoggerFieldValue]:
        return self._key_map

    def lookup_value(self, cls: ta.Type[TypedLoggerValue]) -> ta.Optional[TypedLoggerValueOrProviderOrAbsent]:
        try:
            return self._self_value_map[cls]
        except KeyError:
            pass
        return self._parent.lookup_value(cls)

    @property
    def const_value_map(self) -> ta.Mapping[ta.Type[TypedLoggerValue], TypedLoggerValueOrAbsent]:
        return self._const_value_map

    @property
    def value_wrapper_fn(self) -> ta.Optional[TypedLoggerValueWrapperFn]:
        return self._parent.value_wrapper_fn

    #

    def _typed_logger_visit_bindings(self, vst: 'TypedLoggerBindings._Visitor') -> None:
        self._parent._typed_logger_visit_bindings(vst)  # noqa


##


class UnhandledTypedValueWrapperTypeError(TypeError):
    pass


@ta.final
class TypedLoggerValueWrapper(ta.NamedTuple):
    tys: ta.AbstractSet[type]
    fn: TypedLoggerValueWrapperFn

    #

    def _typed_logger_visit_bindings(self, vst: TypedLoggerBindings._Visitor) -> None:  # noqa
        vst.accept_value_wrapping(self)

    def _typed_logger_visit_value_wrappers(self, fn: ta.Callable[[type, TypedLoggerValueWrapperFn], None]) -> None:  # noqa
        for ty in self.tys:
            fn(ty, self.fn)


##


TYPED_LOGGER_BINDING_ITEM_TYPES: ta.Tuple[ta.Type[TypedLoggerBindingItem], ...] = (
    TypedLoggerField,
    *TYPED_LOGGER_VALUE_OR_PROVIDER_TYPES,
    TypedLoggerBindings,
    TypedLoggerValueWrapper,
)


##


CanTypedLoggerBinding = ta.Union[  # ta.TypeAlias  # omlish-amalg-typing-no-move
    TypedLoggerBindingItem,
    ta.Type[TypedLoggerValue],
    ta.Tuple[
        str,
        ta.Union[
            TypedLoggerFieldValue,
            ta.Any,
        ],
    ],
    None,
]

CanChainTypedLoggerBinding = ta.Union[  # ta.TypeAlias  # omlish-amalg-typing-no-move
    ChainTypedLoggerBindingItem,
    ta.Type[TypedLoggerValue],
    ta.Tuple[
        str,
        ta.Union[
            TypedLoggerFieldValue,
            ta.Any,
        ],
    ],
    None,
]


_AS_TYPED_LOGGER_BINDINGS_DIRECT_TYPES: ta.Tuple[type, ...] = (
    TypedLoggerField,
    TypedLoggerBindings,
    TypedLoggerValueWrapper,
)

_AS_TYPED_LOGGER_BINDINGS_FIELD_VALUE_DIRECT_TYPES: ta.Tuple[type, ...] = (
    *TYPED_LOGGER_VALUE_OR_PROVIDER_OR_ABSENT_TYPES,
    TypedLoggerConstFieldValue,
)


# @ta.final
# class TypedLoggerBindingsBuilder:
#     def __init__(
#             self,
#             *,
#             add_default_keys: bool = False,
#             default_key_filter: ta.Optional[ta.Callable[[str], bool]] = None,
#
#             add_default_values: bool = False,
#             default_value_filter: ta.Optional[ta.Callable[[ta.Type[DefaultTypedLoggerValue]], bool]] = None,
#
#             value_wrapper: ta.Optional[TypedLoggerValueWrapperFn] = None,
#     ) -> None:
#         self._add_default_keys = add_default_keys
#         self._default_key_filter = default_key_filter
#
#         self._add_default_values = add_default_values
#         self._default_value_filter = default_value_filter
#
#         self._value_wrapper = value_wrapper
#
#         self._lst: ta.List[TypedLoggerBindingItem] = []


def as_typed_logger_bindings(
        *objs: CanTypedLoggerBinding,

        add_default_keys: bool = False,
        default_key_filter: ta.Optional[ta.Callable[[str], bool]] = None,

        add_default_values: bool = False,
        default_value_filter: ta.Optional[ta.Callable[[ta.Type[DefaultTypedLoggerValue]], bool]] = None,

        value_wrapper: ta.Optional[TypedLoggerValueWrapperFn] = None,
) -> ta.Sequence[TypedLoggerBindingItem]:
    """This functionality is combined to preserve final key ordering."""

    lst: ta.List[TypedLoggerBindingItem] = []

    for o in objs:
        if o is None:
            continue

        elif isinstance(o, _AS_TYPED_LOGGER_BINDINGS_DIRECT_TYPES):
            lst.append(o)  # type: ignore[arg-type]

        elif isinstance(o, TypedLoggerValue):
            lst.append(o)

            if add_default_keys:
                if (dk := o.default_key()) is not None:
                    if default_key_filter is None or default_key_filter(dk):
                        lst.append(TypedLoggerField(dk, o))

        elif isinstance(o, TypedLoggerValueProvider):
            lst.append(o)

            if add_default_keys:
                if (dk := o.cls.default_key()) is not None:
                    if default_key_filter is None or default_key_filter(dk):
                        lst.append(TypedLoggerField(dk, o))

        elif isinstance(o, type) and issubclass(o, TypedLoggerValue):
            b = False

            if add_default_values and issubclass(o, DefaultTypedLoggerValue):
                if (dp := o.default_provider()) is not None:
                    if default_value_filter is None or default_value_filter(o):
                        lst.append(dp)
                        b = True

            if add_default_keys:
                if (dk := o.default_key()) is not None:
                    if default_key_filter is None or default_key_filter(dk):
                        lst.append(TypedLoggerField(dk, o))
                        b = True

            if not b:
                raise TypeError(f'{o} was added as neither a default key nor a default value')

        elif isinstance(o, tuple):
            k, v = o
            if not isinstance(k, str):
                raise TypeError(k)

            if (
                    isinstance(v, _AS_TYPED_LOGGER_BINDINGS_FIELD_VALUE_DIRECT_TYPES) or
                    (isinstance(o, type) and issubclass(o, TypedLoggerValue))  # type: ignore[unreachable]
            ):
                lst.append(TypedLoggerField(k, v))  # type: ignore[arg-type]

            else:
                lst.append(TypedLoggerField(k, TypedLoggerConstFieldValue(v)))

        elif value_wrapper is not None:
            wv = value_wrapper(o)
            if not isinstance(wv, TypedLoggerValue):
                raise TypeError(wv)

            lst.append(wv)

            if add_default_keys:
                if (dk := wv.default_key()) is not None:
                    if default_key_filter is None or default_key_filter(dk):
                        lst.append(TypedLoggerField(dk, wv))

        else:
            # ta.assert_never(o)
            raise TypeError(o)

    return lst



import dataclasses as dc
import typing as ta


##


@dc.dataclass(frozen=True, kw_only=True)
class PackageConfig:
    codegen: bool = False

    def __init_subclass__(cls, **kwargs):
        raise TypeError


DEFAULT_PACKAGE_CONFIG = PackageConfig()


##


def init_package(
        init_globals: ta.MutableMapping[str, ta.Any],
        *,
        codegen: bool = False,
) -> None:
    if init_globals['__name__'] != init_globals['__package__']:
        raise NameError('Must call dataclasses.init_package from __init__')


##


class PackageConfigCache:
    def __init__(self) -> None:
        super().__init__()

        self._dct: dict[str, PackageConfig | None] = {}

    @ta.overload
    def get(self, pkg: str, default: PackageConfig) -> PackageConfig:
        ...

    @ta.overload
    def get(self, pkg: str, default: PackageConfig | None = None) -> PackageConfig | None:
        ...

    def get(self, pkg, default=None):
        # try:
        #     return self._dct[pkg]
        # except KeyError:
        #     pass
        #
        # try:
        #     s = importlib.resources.read_text(pkg, PACKAGE_CONFIG_FILE_NAME)
        # except FileNotFoundError:
        #     self._dct[pkg] = None
        # else:
        #     c = PackageConfig.loads(s)
        #     self._dct[pkg] = c
        #     return c
        #
        # if '.' not in pkg:
        #     return default
        #
        # return self.get(pkg.rpartition('.')[0], default)
        return None


PACKAGE_CONFIG_CACHE = PackageConfigCache()



import importlib.util
import typing as ta


##


def lazy_import(
        name: str,
        package: str | None = None,
        *,
        optional: bool = False,
        cache_failure: bool = False,
) -> ta.Callable[[], ta.Any]:
    result = not_set = object()

    def inner():
        nonlocal result

        if result is not not_set:
            if isinstance(result, Exception):
                raise result
            return result

        try:
            mod = importlib.import_module(name, package=package)

        except Exception as e:
            if optional:
                if cache_failure:
                    result = None
                return None

            if cache_failure:
                result = e
            raise

        result = mod
        return mod

    return inner



"""
TODO:
 - ContextVar ('context')
 - greenlet?
 - dynamic? https://github.com/wrmsr/iceworm/blob/2f6b4d5e9d237ef9665f7d57cfa6ce328efa0757/iceworm/utils/inject.py#L44
"""
import abc
import contextlib
import threading
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from ..bindings import Binding
from ..elements import Elements
from ..elements import as_elements
from ..errors import ScopeAlreadyOpenError
from ..errors import ScopeNotOpenError
from ..injector import AsyncInjector
from ..keys import Key
from ..providers import FnProvider
from ..providers import Provider
from ..scopes import ScopeSeededProvider
from ..scopes import SeededScope
from ..scopes import Singleton
from ..scopes import ThreadScope
from ..types import Scope
from ..types import Unscoped
from .bindings import BindingImpl
from .providers import ProviderImpl


if ta.TYPE_CHECKING:
    from . import injector as _injector
else:
    _injector = lang.proxy_import('.injector', __package__)


##


class ScopeImpl(lang.Abstract):
    @property
    @abc.abstractmethod
    def scope(self) -> Scope:
        raise NotImplementedError

    def auto_elements(self) -> Elements | None:
        return None

    @abc.abstractmethod
    def provide(self, binding: BindingImpl, injector: AsyncInjector) -> ta.Awaitable[ta.Any]:
        raise NotImplementedError


class UnscopedScopeImpl(ScopeImpl, lang.Final):
    @property
    def scope(self) -> Unscoped:
        return Unscoped()

    async def provide(self, binding: BindingImpl, injector: AsyncInjector) -> ta.Any:
        return await binding.provider.provide(injector)


class SingletonScopeImpl(ScopeImpl, lang.Final):
    def __init__(self) -> None:
        super().__init__()

        self._dct: dict[BindingImpl, ta.Any] = {}

    @property
    def scope(self) -> Singleton:
        return Singleton()

    async def provide(self, binding: BindingImpl, injector: AsyncInjector) -> ta.Any:
        try:
            return self._dct[binding]
        except KeyError:
            pass
        v = await binding.provider.provide(injector)
        self._dct[binding] = v
        return v


class ThreadScopeImpl(ScopeImpl, lang.Final):
    def __init__(self) -> None:
        super().__init__()

        self._local = threading.local()

    @property
    def scope(self) -> ThreadScope:
        return ThreadScope()

    async def provide(self, binding: BindingImpl, injector: AsyncInjector) -> ta.Any:
        dct: dict[BindingImpl, ta.Any]
        try:
            dct = self._local.dct
        except AttributeError:
            dct = self._local.dct = {}
        try:
            return dct[binding]
        except KeyError:
            pass
        v = await binding.provider.provide(injector)
        dct[binding] = v
        return v


##


@dc.dataclass(frozen=True, eq=False)
class ScopeSeededProviderImpl(ProviderImpl):
    p: ScopeSeededProvider

    @property
    def providers(self) -> ta.Iterable[Provider]:
        return (self.p,)

    async def provide(self, injector: AsyncInjector) -> ta.Any:
        ii = check.isinstance(injector, _injector.AsyncInjectorImpl)
        ssi = check.isinstance(ii.get_scope_impl(self.p.ss), SeededScopeImpl)
        return ssi.must_state().seeds[self.p.key]


class SeededScopeImpl(ScopeImpl):
    @dc.dataclass(frozen=True)
    class State:
        seeds: dict[Key, ta.Any]
        prvs: dict[BindingImpl, ta.Any] = dc.field(default_factory=dict)

    def __init__(self, ss: SeededScope) -> None:
        super().__init__()

        self._ss = check.isinstance(ss, SeededScope)
        self._st: SeededScopeImpl.State | None = None

    @property
    def scope(self) -> SeededScope:
        return self._ss

    def must_state(self) -> 'SeededScopeImpl.State':
        if (st := self._st) is None:
            raise ScopeNotOpenError(self._ss)
        return st

    class Manager(SeededScope.Manager, lang.Final):
        def __init__(self, ss: SeededScope, i: AsyncInjector) -> None:
            super().__init__()

            self._ss = check.isinstance(ss, SeededScope)
            self._ii = check.isinstance(i, _injector.AsyncInjectorImpl)
            self._ssi = check.isinstance(self._ii.get_scope_impl(self._ss), SeededScopeImpl)

        @contextlib.asynccontextmanager
        async def __call__(self, seeds: ta.Mapping[Key, ta.Any]) -> ta.AsyncGenerator[None]:
            try:
                if self._ssi._st is not None:  # noqa
                    raise ScopeAlreadyOpenError(self._ss)
                self._ssi._st = SeededScopeImpl.State(dict(seeds))  # noqa
                await self._ii._instantiate_eagers(self._ss)  # noqa
                yield
            finally:
                if self._ssi._st is None:  # noqa
                    raise ScopeNotOpenError(self._ss)
                self._ssi._st = None  # noqa

    def auto_elements(self) -> Elements:
        return as_elements(
            Binding(
                Key(SeededScope.Manager, tag=self._ss),
                FnProvider(lang.typed_partial(SeededScopeImpl.Manager, ss=self._ss)),
                scope=Singleton(),
            ),
        )

    async def provide(self, binding: BindingImpl, injector: AsyncInjector) -> ta.Any:
        st = self.must_state()
        try:
            return st.prvs[binding]
        except KeyError:
            pass
        v = await binding.provider.provide(injector)
        st.prvs[binding] = v
        return v


##


SCOPE_IMPLS_BY_SCOPE: dict[type[Scope], ta.Callable[..., ScopeImpl]] = {
    Unscoped: lambda _: UnscopedScopeImpl(),
    Singleton: lambda _: SingletonScopeImpl(),
    ThreadScope: lambda _: ThreadScopeImpl(),
    SeededScope: lambda s: SeededScopeImpl(s),
}


def make_scope_impl(s: Scope) -> ScopeImpl:
    try:
        fac = SCOPE_IMPLS_BY_SCOPE[type(s)]
    except KeyError:
        pass
    else:
        return fac(s)

    raise TypeError(s)
